"""
    Class for the creation of a scheduler in main.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import threading
import time
import traceback
import typing
from datetime import datetime as dt

from blankly.utils.time_builder import time_interval_to_seconds
from blankly.utils.utils import ceil_date
from blankly.utils.utils import info_print


class Scheduler:
    def __init__(self, function: typing.Callable,
                 interval: typing.Union[str, float],
                 initially_stopped: bool = False,
                 synced: bool = False, **kwargs):
        """
        Wrapper for functions that run at a set interval
        Args:
            function: Function reference to create the scheduler on ex: self.price_event
            interval: int of delay between calls in seconds, or a string that takes units s, m, h, d w, M, y (second,
              minute, hour, day, week, month, year) after a magnitude. examples: "4s", "6h", "10d".
            initially_stopped: Keep the scheduler halted until start() is run
            synced: Align the scheduler with intervals in UTC. ex: if the interval is '1h' then with sync it will only
              run at *:00
        """
        if isinstance(interval, str):
            interval = time_interval_to_seconds(interval)

        # Stop flag
        self.__stop = False
        self.__thread = None
        self.synced = synced

        self.__interval = interval
        self.__kwargs = kwargs
        self.__callback = function

        self.__lock = threading.Lock()

        if not initially_stopped:
            self.start()

        # This can be used for a statically typed, single-process bot decorator.
        # Leaving this because it probably will be used soon

        # def decorator(function):
        #     @functools.wraps(function)
        #     def wrapper():
        #         thread = threading.Thread(target=threading_wait, args=(function, interval,))
        #         thread.start()
        #     wrapper()
        # return decorator

    def start(self, force=False):
        """
        Start the scheduler.

        Args:
            force: Override multi-thread protection and create more threads of the scheduler callback
        """
        if self.__thread is None or force:
            self.__thread = threading.Thread(target=self.__threading_wait,
                                             args=(self.__callback, self.__interval, self.__kwargs))
            self.__thread.start()
        else:
            info_print("Scheduler already started and force not enabled...skipping start.")

    def get_interval(self):
        """
        Get the seconds between each scheduler run
        """
        return self.__interval

    def make_daemon(self):
        self.__thread.setDaemon(True)

    def get_kwargs(self):
        """
        Get the custom function argument dictionary
        """
        return self.__kwargs

    def stop_scheduler(self):
        """
        Halt the scheduler loop
        """
        self.__stop = True

    def get_callback(self):
        """
        Get the callback function that the scheduler is running on
        """
        return self.__callback

    def __threading_wait(self, func, interval, kwargs):
        """
        This function is used with the scheduler decorator
        """
        base_time = time.time()
        if self.synced:
            base_time = ceil_date(dt.now(), seconds=interval).timestamp()
            offset = base_time - time.time()
            time.sleep(offset)
            kwargs['ohlcv_time'] = base_time
        while True:
            if self.__stop:
                break
            # This try except is replicated in the backtesting framework
            try:
                func(**kwargs)
            except Exception:
                traceback.print_exc()
            base_time += interval
            if self.synced:
                kwargs['ohlcv_time'] += interval

            # The downside of this is that it keeps the thread running while waiting to stop
            # It's dependent on delay if its more efficient to just check more
            offset = base_time - time.time()
            if offset < 0:
                offset = 0
            time.sleep(offset)



"""
    This simulates a macd strategy running once a day across many stocks
"""

from blankly import Strategy, StrategyState, Interface
from blankly import Alpaca
from blankly.utils import trunc
from blankly.indicators import macd

SHORT_PERIOD = 12
LONG_PERIOD = 26
SIGNAL_PERIOD = 9


def init(symbol, state: StrategyState):
    interface = state.interface
    resolution = state.resolution
    variables = state.variables
    # initialize the historical data
    variables['history'] = interface.history(symbol, 800, resolution, return_as='deque')['close']
    variables['short_period'] = SHORT_PERIOD
    variables['long_period'] = LONG_PERIOD
    variables['signal_period'] = SIGNAL_PERIOD
    variables['has_bought'] = False


def price_event(price, symbol, state: StrategyState):
    interface: Interface = state.interface
    # allow the resolution to be any resolution: 15m, 30m, 1d, etc.
    variables = state.variables

    variables['history'].append(price)
    macd_res, macd_signal, macd_histogram = macd(variables['history'], 
                                                 short_period=variables['short_period'],
                                                 long_period=variables['long_period'],
                                                 signal_period=variables['signal_period'])

    slope_macd = (macd_res[-1] - macd_res[-5]) / 5  # get the slope of the last 5 MACD_points
    prev_macd = macd_res[-2]
    curr_macd = macd_res[-1]
    curr_signal_macd = macd_signal[-1]

    # We want to make sure this works even if curr_macd does not equal the signal MACD
    is_cross_up = slope_macd > 0 and curr_macd >= curr_signal_macd > prev_macd

    is_cross_down = slope_macd < 0 and curr_macd <= curr_signal_macd < prev_macd
    if is_cross_up:
        # If there is a buy signal, buy with 40% of cash available (that 40% has to be more than 10 dollars though)
        cash = trunc(interface.cash * .4, 2)
        if cash > 10:
            interface.market_order(symbol, 'buy', int(cash/price))
            variables['has_bought'] = True
    elif is_cross_down and variables['has_bought']:
        # Sell all of the position. We also have to own a position after buying
        interface.market_order(symbol, 'sell', int(interface.account[symbol].available))
        variables['has_bought'] = False


if __name__ == "__main__":
    alpaca = Alpaca()
    s = Strategy(alpaca)

    # Add a bunch of stocks to watch for 2 years
    s.add_price_event(price_event, 'SNAP', resolution='1d', init=init)
    s.add_price_event(price_event, 'PBFX', resolution='1d', init=init)
    s.add_price_event(price_event, 'NCLH', resolution='1d', init=init)

    s.add_price_event(price_event, 'CRBP', resolution='1d', init=init)
    s.add_price_event(price_event, 'D', resolution='1d', init=init)
    s.add_price_event(price_event, 'GME', resolution='1d', init=init)

    print(s.backtest(initial_values={'USD': 10000}, to='2y'))



"""
    Custom alpaca interface unit tests
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from pathlib import Path

import pytest
from pytest_mock import MockerFixture

import blankly


class mock_alpaca_direct_calls:
    def __init__(self, *args, **kwargs):
        pass

    def get_account(self):
        mock_account_response = {
            "account_blocked": False,
            "account_number": "010203ABCD",
            "buying_power": "262113.632",
            "cash": "1500",
            "created_at": "2019-06-12T22:47:07.99658Z",
            "currency": "USD",
            "daytrade_count": 0,
            "daytrading_buying_power": "262113.632",
            "equity": "103820.56",
            "id": "e6fe16f3-64a4-4921-8928-cadf02f92f98",
            "initial_margin": "63480.38",
            "last_equity": "103529.24",
            "last_maintenance_margin": "38000.832",
            "long_market_value": "126960.76",
            "maintenance_margin": "38088.228",
            "multiplier": "4",
            "pattern_day_trader": False,
            "portfolio_value": "103820.56",
            "regt_buying_power": "80680.36",
            "short_market_value": "0",
            "shorting_enabled": True,
            "sma": "0",
            "status": "ACTIVE",
            "trade_suspended_by_user": False,
            "trading_blocked": False,
            "transfers_blocked": False
        }
        return mock_account_response

    def list_orders(self, *args, **kwargs):
        mock_orders_response = [{
            "id": "61e69015-8549-4bfd-b9c3-01e75843f47d",
            "client_order_id": "eb9e2aaa-f71a-4f51-b5b4-52a6c565dad4",
            "created_at": "2021-03-16T18:38:01.942282Z",
            "updated_at": "2021-03-16T18:38:01.942282Z",
            "submitted_at": "2021-03-16T18:38:01.937734Z",
            "filled_at": None,
            "expired_at": None,
            "canceled_at": None,
            "failed_at": None,
            "replaced_at": None,
            "replaced_by": None,
            "replaces": None,
            "asset_id": "b0b6dd9d-8b9b-48a9-ba46-b9d54906e415",
            "symbol": "AAPL",
            "asset_class": "us_equity",
            "notional": None,
            "qty": 1,
            "filled_qty": "0",
            "filled_avg_price": None,
            "order_class": "",
            "order_type": "market",
            "type": "market",
            "side": "sell",
            "time_in_force": "day",
            "limit_price": None,
            "stop_price": None,
            "status": "accepted",
            "extended_hours": False,
            "legs": None,
            "trail_percent": None,
            "trail_price": None,
            "hwm": None
        }]

        return mock_orders_response

    def get_snapshots(self, symbols):
        mock_snapshots = [{
            "symbol": "AAPL",
            "latestTrade": {
                "t": "2021-05-11T20:00:00.435997104Z",
                "x": "Q",
                "p": 125.91,
                "s": 5589631,
                "c": [
                    "@",
                    "M"
                ],
                "i": 179430,
                "z": "C"
            },
            "latestQuote": {
                "t": "2021-05-11T22:05:02.307304704Z",
                "ax": "P",
                "ap": 125.68,
                "as": 12,
                "bx": "P",
                "bp": 125.6,
                "bs": 4,
                "c": [
                    "R"
                ]
            },
            "minuteBar": {
                "t": "2021-05-11T22:02:00Z",
                "o": 125.66,
                "h": 125.66,
                "l": 125.66,
                "c": 125.66,
                "v": 396
            },
            "dailyBar": {
                "t": "2021-05-11T04:00:00Z",
                "o": 123.5,
                "h": 126.27,
                "l": 122.77,
                "c": 125.91,
                "v": 125863164
            },
            "prevDailyBar": {
                "t": "2021-05-10T04:00:00Z",
                "o": 129.41,
                "h": 129.54,
                "l": 126.81,
                "c": 126.85,
                "v": 79569305
            }
        }]

        return mock_snapshots

    def list_positions(self):
        mock_positions_response = [{
            "asset_id": "904837e3-3b76-47ec-b432-046db621571b",
            "symbol": "AAPL",
            "exchange": "NASDAQ",
            "asset_class": "us_equity",
            "avg_entry_price": "100.0",
            "qty": "5",
            "side": "long",
            "market_value": "600.0",
            "cost_basis": "500.0",
            "unrealized_pl": "100.0",
            "unrealized_plpc": "0.20",
            "unrealized_intraday_pl": "10.0",
            "unrealized_intraday_plpc": "0.0084",
            "current_price": "120.0",
            "lastday_price": "119.0",
            "change_today": "0.0084"
        }]
        return mock_positions_response

    def list_assets(self, *args, **kwargs):
        mock_products_response = [{
            "id": "904837e3-3b76-47ec-b432-046db621571b",
            "class": "us_equity",
            "exchange": "NASDAQ",
            "symbol": "AAPL",
            "status": "active",
            "tradable": True,
            "marginable": True,
            "shortable": True,
            "easy_to_borrow": True,
            "fractionable": True
        }]
        return mock_products_response


class mock_alpaca_interface:
    def __init__(self, *args, **kwargs):
        pass

    def get_account(self):
        return "hello"


@pytest.fixture
def alpaca_mock_interface(mocker: MockerFixture):
    keys_file_path = Path("tests/config/keys.json").resolve().__str__()
    settings_file_path = Path("tests/config/settings.json").resolve().__str__()

    mocker.patch("alpaca_trade_api.REST", new=mock_alpaca_direct_calls)
    alpaca = blankly.Alpaca(keys_path=keys_file_path,
                            settings_path=settings_file_path,
                            portfolio_name="alpaca test portfolio")
    # auth_obj = AlpacaAuth(keys_file_path, "alpaca test portfolio")
    # _, alpaca_interface = DirectCallsFactory.create("alpaca", auth_obj, settings_file_path)

    return alpaca.interface


# TODO: Need to create a functional testing package
# @pytest.fixture
# def alpaca_interface(mocker: MockerFixture) -> None:
#     keys_file_path = Path("tests/config/keys.json").resolve()
#     settings_file_path = Path("tests/config/settings.json").resolve()
#
#     auth_obj = alpaca_auth(keys_file_path, "alpaca test portfolio")
#     alpaca_interface = DirectCallsFactory.create_interface("alpaca", auth_obj, settings_file_path)
#     return alpaca_interface
#
#
# def test_alpaca_interface_functional(alpaca_interface) -> None:
#
#     assert alpaca_interface.get_exchange_type() == 'alpaca'


# TODO: make this test better
def test_get_account(alpaca_mock_interface) -> None:
    return_val = alpaca_mock_interface.get_account()

    expected_answer = {
        "AAPL": {
            "available": 4,
            "hold": 1,
        }
    }

    assert 'AAPL' in return_val
    assert return_val['AAPL']['available'] == 4
    assert return_val['AAPL']['hold'] == 1


def test_get_cash(alpaca_mock_interface) -> None:
    expected_answer = 262113.632
    return_val = alpaca_mock_interface.cash
    assert return_val == expected_answer


def test_get_fees(alpaca_mock_interface) -> None:
    fee_response = alpaca_mock_interface.get_fees()
    assert fee_response['maker_fee_rate'] == 0
    assert fee_response['taker_fee_rate'] == 0



from blankly.indicators.indicators import *
from blankly.indicators.moving_averages import *
from blankly.indicators.oscillators import *
from blankly.indicators.statistics import *
from blankly.indicators.utils import *



"""
    Abstraction for websocket manager objects
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import blankly.utils.utils
from blankly.exchanges.abc_exchange_websocket import ABCExchangeWebsocket


class WebsocketManager(ABCExchangeWebsocket):
    def __init__(self, websockets, default_symbol, default_exchange):
        self.websockets = websockets
        self.__default_symbol = default_symbol
        self.__default_exchange = default_exchange

        self.preferences = blankly.utils.load_user_preferences()

    def close_all_websockets(self):
        """
        Iterate through orderbooks and make sure they're closed
        """
        self.__recursive_close(self.websockets)

    def __recursive_close(self, d):
        for k, v in d.items():
            if isinstance(v, dict):
                self.__recursive_close(v)
            else:
                d[k].close_websocket()

    def __evaluate_overrides(self, override_symbol, override_exchange):
        """
        Switches to inputted exchanges, used in the public methods
        """
        # Issues could arise if the user specifies a different exchange but not a different currency. It would run
        # that default on the new exchange. Not great behavior unless they're clever
        if override_symbol is not None:
            currency_id = override_symbol
        else:
            currency_id = self.__default_symbol
        if override_exchange is not None:
            exchange = override_exchange
        else:
            exchange = self.__default_exchange

        if exchange == "binance":
            currency_id = blankly.utils.to_exchange_symbol(currency_id, "binance")
        if exchange == "ftx":
            currency_id = blankly.utils.to_exchange_symbol(currency_id, "ftx")
        return self.websockets[exchange][currency_id]

    def get_ticker(self, symbol, override_exchange=None):
        """
        Retrieve the ticker attached to a currency
        Args:
            symbol: The currency ID of ticker (required) such as "BTC-USD"
            override_exchange: Override the default to get tickers for different exchanges
        """
        if override_exchange is None:
            return self.websockets[self.__default_exchange][symbol]
        else:
            return self.websockets[override_exchange][symbol]

    def get_all_tickers(self) -> dict:
        """
        Get the tickers object dictionary. This can be used for individual management, or to stop using the manager.
        Returns:
            Dictionary with tickers referenced for each currency
        """
        return self.websockets

    """
    Ticker functions/overridden
    """

    def append_callback(self, callback_object, override_symbol=None, override_exchange=None):
        """
        This bypasses all processing that a manager class may do before returning to the user's main.
        For example with an orderbook feed, this will return the ticks instead of a sorted orderbook
        This can be very useful in working with multiple tickers, but not necessary on a simple bot.

        Args:
            callback_object: Reference for the callback function. The price_event(self, tick)
                function would be passed in as just self.price_event  -- no parenthesis or arguments, just the object
            override_symbol: Ticker id, such as "BTC-USD" or exchange equivalents.
            override_exchange: Forces the manager to use a different supported exchange.
        """
        websocket = self.__evaluate_overrides(override_symbol, override_exchange)

        websocket.append_callback(callback_object)

    def is_websocket_open(self, override_symbol=None, override_exchange=None) -> bool:
        """
        Check if the websocket attached to a currency is open
        """
        websocket = self.__evaluate_overrides(override_symbol, override_exchange)

        return websocket.is_websocket_open()

    def get_most_recent_time(self, override_symbol=None, override_exchange=None):
        """
        Get the most recent time associated with the most recent tick
        """
        websocket = self.__evaluate_overrides(override_symbol, override_exchange)

        return websocket.get_most_recent_time()

    def get_time_feed(self, override_symbol=None, override_exchange=None):
        """
        Get a time array associated with the ticker feed.
        """
        websocket = self.__evaluate_overrides(override_symbol, override_exchange)

        return websocket.get_time_feed()

    def get_feed(self, override_symbol=None, override_exchange=None):
        """
        Get the full ticker array. This can be extremely large.
        """
        websocket = self.__evaluate_overrides(override_symbol, override_exchange)

        return websocket.get_feed()

    def get_response(self, override_symbol=None, override_exchange=None):
        """
        Get the exchange's response to the request to subscribe to a feed
        """
        websocket = self.__evaluate_overrides(override_symbol, override_exchange)

        return websocket.get_response()

    def close_websocket(self, override_symbol=None, override_exchange=None):
        """
        Close a websocket thread
        """
        websocket = self.__evaluate_overrides(override_symbol, override_exchange)

        websocket.close_websocket()

    def restart_ticker(self, override_symbol=None, override_exchange=None):
        """
        Restart a websocket feed after asking it to stop
        """
        websocket = self.__evaluate_overrides(override_symbol, override_exchange)

        websocket.restart_ticker()

    def get_most_recent_tick(self, override_symbol=None, override_exchange=None):
        """
        Get the most recent tick received
        """
        websocket = self.__evaluate_overrides(override_symbol, override_exchange)
        # TODO fix the returned value below, really this needs a class like in binance that can create a callback to
        #  allow a pointer to be subbed in for whichever exchange/currency/websocket type is overridden
        return websocket.get_most_recent_tick()



"""
    Unit tests for Coinbase Pro.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import unittest
from pathlib import Path

import pytest

import blankly
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro_interface import CoinbaseProInterface


class CoinbaseInterface2(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.Coinbase_Pro = blankly.CoinbasePro(portfolio_name="Sandbox Portfolio",
                                               keys_path='./tests/config/keys.json',
                                               settings_path="./tests/config/settings.json")
        cls.Coinbase_Pro_Interface = cls.Coinbase_Pro.get_interface()

    def test_get_exchange_type(self):
        coinbase_pro = self.Coinbase_Pro_Interface.get_exchange_type()
        self.assertEqual(coinbase_pro, "coinbase_pro")


@pytest.fixture
def coinbase_interface():
    keys_file_path = Path("tests/config/keys.json").resolve()
    settings_file_path = Path("tests/config/settings.json").resolve()

    coinbase_pro = blankly.CoinbasePro(keys_path=keys_file_path,
                                       settings_path=settings_file_path,
                                       portfolio_name='Sandbox Portfolio')

    # auth_obj = CoinbaseProAuth(str(keys_file_path), "Sandbox Portfolio")
    # _, coinbase_interface = DirectCallsFactory.create("coinbase_pro", auth_obj, str(settings_file_path))
    return coinbase_pro.interface


def test_get_exchange(coinbase_interface: CoinbaseProInterface) -> None:
    assert coinbase_interface.get_exchange_type() == 'coinbase_pro'

    products = coinbase_interface.get_products()
    btc_usd_id = None

    for product in products:
        if product['base_asset'] == 'BTC' and product['quote_asset'] == 'USD':
            btc_usd_id = product['symbol']

    assert btc_usd_id

    # market_buy_order = coinbase_interface.market_order(btc_usd_id, 'buy', 200)


def test_get_price(coinbase_interface: CoinbaseProInterface) -> None:
    responses = []
    responses.append(coinbase_interface.get_price('BTC-USD'))
    responses.append(coinbase_interface.get_price('BTC-EUR'))
    responses.append(coinbase_interface.get_price('BTC-GBP'))
    responses.append(coinbase_interface.get_price('ETH-BTC'))
    for resp in responses:
        assert type(resp) is float


# TODO This test is removed because it overloads simultaneous history tests
# def test_start_with_end_history(coinbase_interface: CoinbaseProInterface) -> None:
#     # This initial selection could fail because of the slightly random day that they delete their data
#     stop_dt = dateparser.parse("2021-08-04")
#     start = "2021-01-04"
#     stop = str(stop_dt.date())
#
#     # The dates are offset by one because the time is the open time
#     close_stop = str(stop_dt.replace(day=stop_dt.day - 1).date())
#
#     resp = coinbase_interface.history('BTC-USD', resolution='1h', start_date=start, end_date=stop)
#
#     start_date = dt.fromtimestamp(resp['time'][0]).strftime('%Y-%m-%d')
#     end_date = dt.fromtimestamp(resp['time'].iloc[-1]).strftime('%Y-%m-%d')
#
#     assert start_date == start
#     assert end_date == close_stop



"""
    Screener management system for starting & stopping long term monitoring
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import typing
from typing import List
from blankly.frameworks.screener.screener_runner import ScreenerRunner
from blankly.utils.utils import load_deployment_settings

import blankly
from blankly.exchanges.exchange import Exchange
from blankly.frameworks.screener.screener_state import ScreenerState
from copy import deepcopy


class Screener:
    def __init__(self, exchange: Exchange,
                 evaluator: typing.Callable,
                 symbols: List[str],
                 init: typing.Callable = None,
                 final: typing.Callable = None,
                 formatter: typing.Callable = None):
        """
        Create a new screener.

        Function Signatures:
        init(screener_state: blankly.ScreenerState)
        evaluator(symbol: str, screener_state: blankly.ScreenerState)
        final(ScreenerState: blankly.ScreenerState)
        formatter(raw_results: dict, screener_state: blankly.ScreenerState)

        This heavily differs from Strategy objects. While a Strategy is optimized for the implementation of
         short or long-term trading strategies, a Screener is optimized for long-term monitoring & reporting of many
         symbols. Screeners are designed to be scheduled to run over intervals of days & weeks. When deployed live,
         a screener-based script will only start when scheduled, then exit entirely.

        Args:
            exchange: An exchange object to construct the screener on
            evaluator: The function that can take information about a screener & classify that screener based on
             parameters
            symbols: A list of symbols to run on.
            init: Optional setup code to run when the program starts
            final: Optional teardown code to run before the program finishes. This will be run every time the
             screener finishes a cycle
            formatter: Optional formatting function that pretties the results form the evaluator
        """

        if not blankly.is_deployed and blankly._screener_runner is None:
            cron_settings = load_deployment_settings()['screener']['schedule']
            blankly._screener_runner = ScreenerRunner(cron_settings)

        # TODO export the symbols here as a list
        self.exchange = exchange
        self.symbols = symbols

        # Store callables as a dictionary
        self.__callables = {
            'evaluator': evaluator,
            'init': init,
            'teardown': final,
            'formatter': formatter
        }
        self.interface = exchange.interface

        # Creat the screener state and pass in this screener object
        self.screener_state = ScreenerState(self)

        self.raw_results = {}
        self.formatted_results = {}

        # Note that only a single screener can be exported at a time for a model
        if blankly.is_deployed:
            blankly.reporter.export_screener(self)

        self.__run()

    def __run(self):
        init = self.__callables['init']
        if callable(init):
            init(self.screener_state)

        self.symbols = self.screener_state.symbols

        # Evaluate using the evaluator function
        # 'symbol': {
        #     # This can all be custom
        #     'classification': 'pass/fail or True / False'
        #     'notes': {
        #         'custom_note': 'this stock was cool'
        #         'current_rsi': .45
        #     }
        # }

        evaluator = self.__callables['evaluator']
        if not callable(evaluator):
            raise TypeError("Must pass a callable for the evaluator.")

        for i in self.symbols:
            self.raw_results[i] = evaluator(i, self.screener_state)

        self.symbols = self.screener_state.symbols

        # Copy the evaluator results so that they can be formatted
        self.formatted_results = deepcopy(self.raw_results)

        formatter = self.__callables['formatter']
        if callable(formatter):
            # Mutate the copied dictionary
            self.formatted_results = formatter(self.formatted_results, self.screener_state)

        self.symbols = self.screener_state.symbols

        teardown = self.__callables['teardown']
        if callable(teardown):
            teardown(self.screener_state)

        self.symbols = self.screener_state.symbols

        blankly.reporter.export_screener_result(self)

    def notify(self, message: str = None):
        """
        Send an email and text message to yourself. When deployed live this will come from an official blankly email &
         phone number. When run locally it will use information found in your notify.json file.
         **Note that you need the SMTP configuration to send text messages

        Args:
            message: Optionally fill this with a different string to notify with. If not filled it will notify using the
             formatted results evaluated on construction
        """
        use_str = self.formatted_results
        if message is not None:
            use_str = message

        blankly.reporter.email(use_str)

        blankly.reporter.text(use_str)



"""
    Class to manage a market order lifecycle - a way to determine individual information about each buy and sell
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from blankly.exchanges.orders.order import Order


class MarketOrder(Order):
    """
    (Buying or selling (string), amount in currency (BTC/XLM), ticker object (so we can get time and value),
    limit if there is one)
    """

    def __init__(self, order, response, interface):
        """
        Coinbase Response:
        {
            'id': '5955cd5d-a78c-4c47-9741-1f7f20ac7f95',
            'product_id': 'BTC-USD',
            'side': 'buy',
            'stp': 'dc',
            'funds': '10.97804391',
            'specified_funds': '11',
            'type': 'market',
            'post_only': False,
            'created_at': 1621039946.320796,
            'fill_fees': '0',
            'filled_size': '0',
            'executed_value': '0',
            'status': 'pending',
            'settled': False
        }

        binance Response:
        TODO add this

        Guaranteed:
        needed = [
            ["symbol", str],
            ["id", str],
            ["created_at", float],
            ["funds", float],
            ["status", str],
            ["type", str],
            ["side", str]
        ]
        """
        self.Interface = interface
        self.__exchange = self.Interface.get_exchange_type()
        self.__order = order
        self.__response = response
        self.__homogenized_result = None
        super().__init__(self.__response, self.__order, self.Interface)

    def __str__(self):
        return_string = super().__str__()

        return return_string

    """ This is something that should be implemented in the future """
    # """ Returns true if the buy order can be sold at a profit or the sell order didn't loose money """
    # def is_profitable(self, show=False):
    #     if self.__buyOrSell == "buy":
    #         if self.is_profitable_buy() < float(self.__ticker.get_most_recent_tick()["price"]):
    #             if show:
    #                 print(True)
    #             return True
    #         else:
    #             if show:
    #                 print(False)
    #             return False
    #     else:
    #         if self.is_profitable_buy() > float(self.__ticker.get_most_recent_tick()["price"]):
    #             if show:
    #                 print(True)
    #             return True
    #         else:
    #             if show:
    #                 print(False)
    #             return False
    #
    # def get_fee(self, show=False):
    #     fee = (self.__valueAtTime * self.__amountCurrency) * blankly.Constants.PRETEND_FEE_RATE
    #     if show:
    #         print(fee)
    #     return fee
    #
    # """
    # Returns the price that a BUY needs to reach to be profitable
    # Ths can be used to set limit orders
    # """
    # def is_profitable_buy(self, show=False):
    #     # price = self.__valueAtTime/(1 - Constants.PRETEND_FEE_RATE)
    #     # This one didn't work for the USD amount modified
    #     # price = self.__valueAtTime/(1 - (2 * Constants.PRETEND_FEE_RATE) + (Constants.PRETEND_FEE_RATE * Constants.PRETEND_FEE_RATE))
    #     maker_fee_rate = self.__exchange_properties['maker_fee_rate']
    #     price = ((maker_fee_rate + 1) * (self.__valueAtTime) / (1 - maker_fee_rate))
    #     if show:
    #         print(price)
    #     return price
    #
    # """
    # This allows this to be checked if this buy became profitable
    # """
    # def is_profitable_sell(self, show=False):
    #     self.__profitable = float((self.__ticker.get_most_recent_tick()["price"])) > self.is_profitable_buy()
    #     if show:
    #         print(self.__profitable)
    #     return self.__profitable



"""
    Interface for exchange objects.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import abc


class ABCExchange(abc.ABC):
    """
    Functions required for easy interaction with the GUI/main
    """

    @abc.abstractmethod
    def append_model(self, model, coin, args):
        pass

    @abc.abstractmethod
    def get_full_state(self, symbol):
        pass

    @abc.abstractmethod
    def get_model_state(self, symbol):
        pass

    @abc.abstractmethod
    def get_asset_state(self, symbol):
        pass

    def get_exchange_state(self):
        pass



"""
This is a commentated but cleaner implementation of strategy_base.py

Largely this project is going on the backburner
"""

# """
#     Abstraction for creating interval driven user strategies
#     Copyright (C) 2021  Emerson Dove, Brandon Fan
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Lesser General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
# """
#
# import datetime
# import threading
# import time
# import typing
# import warnings
#
# import pandas as pd
#
# import blankly
# from blankly.exchanges.exchange import Exchange
# from blankly.exchanges.interfaces.paper_trade.backtest_controller import BackTestController
# from blankly.exchanges.interfaces.paper_trade.backtest_result import BacktestResult
# from blankly.frameworks.strategy.strategy_state import StrategyState
# from blankly.utils.time_builder import time_interval_to_seconds
# from blankly.utils.utils import AttributeDict, info_print
# from blankly.utils.utils import get_ohlcv_from_list
# from blankly.exchanges.strategy_logger import StrategyLogger
#
#
# class Strategy:
#     def __init__(self, exchange: Exchange):
#         """
#         Create a new strategy object. A strategy can be used to run your code live while be backtestable and modular
#          across exchanges.
#          Args:
#              exchange: An exchange object. This can be created by doing something similar to exchange = blankly.Alpaca()
#
#         Function Signatures:
#         init(symbol: str, state: blankly.StrategyState)
#         price_event(price: float, symbol: str, state: blankly.StrategyState)
#         orderbook_event(tick: dict, symbol: str, state: blankly.StrategyState)
#         bar_event(bar: dict, symbol: str, state: blankly.StrategyState)
#         teardown(blankly.StrategyState)
#         """
#         self.__exchange = exchange
#         self.__backtest_scheduler: dict = {}
#
#         self.interface = StrategyLogger(interface=exchange.get_interface(), strategy=self)
#
#         self.__schedulers = []
#
#         self.ticker_manager = blankly.TickerManager(self.__exchange.get_type(), '')
#
#     def __idle_event(self, *args, **kwargs):
#         """
#         Function to skip & ignore callbacks. Mainly used for websockets
#         """
#         pass
#
#     def add_price_event(self, callback: typing.Callable, symbol: str, resolution: typing.Union[str, float],
#                         init: typing.Callable = None, teardown: typing.Callable = None, synced: bool = False,
#                         variables: dict = None):
#         """
#         Add Price Event. This will provide you with an updated price every time the callback is run
#         Args:
#             callback: The price event callback that will be added to the current ticker and run at the proper resolution
#             symbol: Currency pair to create the price event for
#             resolution: The resolution that the callback will be run - in seconds
#             init: Callback function to allow a setup for the strategy variable. Example usages include
#                 downloading price data before usage
#             teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
#                 positions, writing or cleaning up data or anything else useful
#             synced: Sync the function to
#             variables: A dictionary to initialize the state's internal values
#         """
#         self.__custom_price_event(callback=callback, symbol=symbol, resolution=resolution, init=init, synced=synced,
#                                   teardown=teardown, variables=variables, type_='price_event')
#
#     def add_bar_event(self, callback: typing.Callable, symbol: str, resolution: typing.Union[str, float],
#                       init: typing.Callable = None, teardown: typing.Callable = None, variables: dict = None):
#         """
#         The bar event sends a dictionary of {open, high, low, close, volume} which has occurred in the interval.
#         Args:
#             callback: The price event callback that will be added to the current ticker and run at the proper resolution
#             symbol: Currency pair to create the price event for
#             resolution: The resolution that the callback will be run - in seconds
#             init: Callback function to allow a setup for the strategy variable. Example usages include
#                 downloading price data before usage
#             teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
#                 positions, writing or cleaning up data or anything else useful
#             variables: A dictionary to initialize the state's internal values
#         """
#         self.__custom_price_event(type_='bar', synced=True, bar=True, callback=callback, symbol=symbol,
#                                   resolution=resolution, init=init, teardown=teardown, variables=variables)
#
#     def __custom_price_event(self,
#                              type_: str,
#                              callback: typing.Callable = None,
#                              symbol: str = None,
#                              resolution: typing.Union[str, float] = None,
#                              init: typing.Callable = None,
#                              synced: bool = False,
#                              bar: bool = False,
#                              teardown: typing.Callable = None, variables: dict = None):
#         """
#         Add Price Event
#         Args:
#             type_: The type of event that the price event refers to
#             callback: The price event callback that will be added to the current ticker and run at the proper
#              resolution
#             symbol: Currency pair to create the price event for
#             resolution: The resolution that the callback will be run - in seconds
#             init: Callback function to allow a setup for the strategy variable. Example usages include
#                 downloading price data before usage
#             teardown: A function to run when the strategy is stopped or interrupted. Example usages include
#              liquidating
#                 positions, writing or cleaning up data or anything else useful
#             synced: Sync the function to
#             bar: Get the OHLCV data for a valid exchange interval
#             variables: Initial dictionary to write into the state variable
#         """
#         # Make sure variables is always an empty dictionary if None
#         if variables is None:
#             variables = {}
#
#         resolution = time_interval_to_seconds(resolution)
#
#         # Make sure the type is always added to the price events
#         if type_ not in self.__backtest_scheduler:
#             self.__backtest_scheduler[type_] = []
#
#         self.__backtest_scheduler[type_] = {
#             'symbol': symbol,
#             'resolution': resolution
#         }
#
#         variables_ = AttributeDict(variables)
#         state = StrategyState(self, variables_, symbol, resolution=resolution)
#
#         scheduler_kwargs = {
#             'initially_stopped': True,
#             'callback': callback,
#             'resolution': resolution,
#             'variables': variables_,
#             'state_object': state,
#             'synced': synced,
#             'init': init,
#             'teardown': teardown,
#             'ohlc': bar,
#             'symbol': symbol
#         }
#
#         if resolution < 10:
#             self.ticker_manager.create_ticker(self.__idle_event, override_symbol=symbol)
#             self.__schedulers.append(
#                 blankly.Scheduler(self.__price_event_websocket, resolution,
#                                   **scheduler_kwargs)
#             )
#             exchange_type = self.__exchange.get_type()
#
#     def __price_event_rest(self, **kwargs):
#         callback = kwargs['callback']
#         symbol = kwargs['symbol']
#         resolution = kwargs['resolution']
#         variables = kwargs['variables']
#         ohlc = kwargs['ohlc']
#         state = kwargs['state_object']  # type: StrategyState
#
#         state.variables = variables
#         state.resolution = resolution
#
#         if ohlc:
#             ohlcv_time = kwargs['ohlcv_time']
#             while True:
#                 # Sometimes coinbase doesn't download recent data correctly
#                 try:
#                     data = self.interface.history(symbol=symbol, to=1, resolution=resolution).iloc[-1].to_dict()
#                     if data['time'] + resolution == ohlcv_time:
#                         break
#                 except IndexError:
#                     pass
#                 time.sleep(.5)
#             data['price'] = self.interface.get_price(symbol)
#         else:
#             data = self.interface.get_price(symbol)
#
#     def __price_event_websocket(self, **kwargs):
#         callback = kwargs['callback']
#         symbol = kwargs['symbol']
#         resolution = kwargs['resolution']
#         variables = kwargs['variables']
#         ohlc = kwargs['ohlc']
#         state = kwargs['state_object']  # type: StrategyState
#
#         if ohlc:
#             close_time = kwargs['ohlcv_time']
#             open_time = close_time-resolution
#             ticker_feed = list(reversed(self.ticker_manager.get_feed(override_symbol=symbol)))
#             #     tick        tick
#             #      |    ohlcv close                            ohlcv open
#             # 0    |   -20          -40            -60        -80
#             # newest, newest - 1, newest - 2
#             count = 0
#             while ticker_feed[count]['time'] > close_time:
#                 count += 1
#
#             close_index = count
#
#             # Start at the close index to save some iterations
#             count = close_index
#             while ticker_feed[count]['time'] < open_time:
#                 count += 1
#             # Subtract 1 to put it back inside the range
#             count -= 1
#             open_index = count
#
#             # Get the latest price that isn't past the timeframe
#             last_price = ticker_feed[close_index:][-1]['price']
#
#             data = get_ohlcv_from_list(list(reversed(ticker_feed[close_index:open_index])), last_price)
#
#         else:
#             try:
#                 data = self.ticker_manager.get_most_recent_tick(override_symbol=symbol)['price']
#             except TypeError:
#                 info_print("No valid data yet - using rest.")
#                 data = self.interface.get_price(symbol)
#
#         state.variables = variables
#         state.resolution = resolution
#
#         callback(data, symbol, state)



"""
    Core logic to reduce duplication across websocket implementations
    Copyright (C) 2022  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import abc
import collections
import threading

import websocket

import blankly.utils.utils
from blankly.exchanges.abc_exchange_websocket import ABCExchangeWebsocket
from blankly.utils.utils import info_print


class Websocket(ABCExchangeWebsocket, abc.ABC):
    def __init__(self, symbol, stream, log, log_message, url, pre_event_callback):
        self.symbol = symbol
        self.stream = stream

        # Initialize log file
        if log is not None:
            self.log = True
            self.file_path = log
            try:
                self.__file = open(log, 'x+')
                self.__file.write(log_message)
            except FileExistsError:
                self.__file = open(log, 'a')
        else:
            self.log = False

        self.url = url

        self.thread = None
        self.message_count = 0

        self.response = None
        self.most_recent_tick = None
        self.most_recent_time = None
        self.callbacks = []
        self.pre_event_callback = pre_event_callback

        # Reload preferences
        self.preferences = blankly.utils.load_user_preferences()
        buffer_size = self.preferences['settings']['websocket_buffer_size']
        self.ticker_feed = collections.deque(maxlen=buffer_size)
        self.time_feed = collections.deque(maxlen=buffer_size)

        self.ws = None

    def start_websocket(self, on_open: callable, on_message: callable, on_error: callable, on_close: callable,
                        target: callable):
        """
        Restart websocket if it was asked to stop.
        """
        if self.ws is None:
            self.ws = websocket.WebSocketApp(self.url,
                                             on_open=on_open,
                                             on_message=on_message,
                                             on_error=on_error,
                                             on_close=on_close)
            self.thread = threading.Thread(target=target)
            self.thread.start()
        else:
            if self.thread.is_alive():
                info_print("Already running...")
            else:
                # Use recursion to restart, continue appending to time feed and ticker feed
                self.ws = None
                self.start_websocket(on_open, on_message, on_error, on_close, target)

    def log_response(self, logging_callback: callable, message: dict):
        # Run callbacks on message
        if self.log:
            if self.message_count % 100 == 0:
                self.__file.close()
                self.__file = open(self.file_path, 'a')
            line = logging_callback(message)
            self.__file.write(line)
    """
    The are access functions
    """
    """ Required in manager """

    def is_websocket_open(self):
        if self.thread is not None:
            return self.thread.is_alive()
        else:
            return False

    def get_currency_id(self):
        return self.symbol

    """ Required in manager """

    def append_callback(self, obj):
        self.callbacks.append(obj)

    """ Define a variable each time so there is no array manipulation """
    """ Required in manager """

    def get_most_recent_tick(self):
        return self.most_recent_tick

    """ Required in manager """

    def get_most_recent_time(self):
        return self.most_recent_time

    """ Required in manager """

    def get_time_feed(self):
        return list(self.time_feed)

    """ Parallel with time feed """
    """ Required in manager """

    def get_feed(self):
        return list(self.ticker_feed)

    """ Required in manager """

    def get_response(self):
        return self.response

    """ Required in manager """

    def close_websocket(self):
        if self.thread is not None and self.thread.is_alive():
            self.ws.close()
        else:
            print("Websocket for " + self.symbol + '@' + self.stream + " is already closed")

    @abc.abstractmethod
    def on_open(self, ws):
        pass

    @abc.abstractmethod
    def on_error(self, ws, error):
        pass

    @abc.abstractmethod
    def on_message(self, ws, message):
        pass

    @abc.abstractmethod
    def on_close(self, ws):
        pass



"""
    General indicators & metrics for blankly
    Copyright (C) 2021  Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    Big thanks to John from Codearmo
"""

import numpy as np
import pandas as pd

from blankly.utils.utils import info_print


def cagr(start_value, end_value, years):
    if end_value < start_value:
        info_print('End Value less than Start Value makes CAGR meaningless, returning 0')
        return 0.0
    return (end_value / start_value) ** (1 / years) - 1


def cum_returns(start_value, end_value):
    return (end_value - start_value) / start_value


def sortino(returns, n=252, risk_free_rate=None):
    returns = pd.Series(returns)
    if risk_free_rate:
        mean = returns.mean() * n - risk_free_rate
    else:
        mean = returns.mean() * n
    std_neg = returns[returns < 0].std() * np.sqrt(n)
    return mean / std_neg


def sharpe(returns, n=252, risk_free_rate=None):
    returns = pd.Series(returns)
    if risk_free_rate:
        mean = returns.mean() * n - risk_free_rate
    else:
        mean = returns.mean() * n
    std = returns.std() * np.sqrt(n)
    return mean / std


def calmar(returns, n=252):
    return_series = pd.Series(returns)
    return return_series.mean() * n / abs(max_drawdown(return_series))


def volatility(returns, n=None):
    return np.std(returns) * np.sqrt(n) if n else np.std(returns)


def variance(returns, n=None):
    return np.nanvar(returns) * n if n else np.nanvar(returns)


def beta(returns, market_base_returns, n=None):
    m = np.matrix([returns, market_base_returns])
    return n * np.cov(m)[0][1] / variance(market_base_returns, n)


def var(initial_value, returns, alpha: float):
    returns_sorted = np.sort(returns)
    index = int(alpha * len(returns_sorted))
    return initial_value * abs(returns_sorted[index])


def cvar(initial_value, returns, alpha):
    returns_sorted = np.sort(returns)
    index = int(alpha * len(returns_sorted))
    sum_var = returns_sorted[0]
    for i in range(1, index):
        sum_var += returns_sorted[i]
    return initial_value * abs(sum_var / index)


def max_drawdown(returns):
    returns = pd.Series(returns)
    cumulative = (returns + 1).cumprod()
    peak = cumulative.expanding(min_periods=1).max()
    dd = (cumulative / peak) - 1
    return dd.min()



"""
    Unit tests for validating crypto behaviors
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import time

import blankly


def test_crypto_websocket():
    test = CryptoWebsockets()

    assert test.passed


class CryptoWebsockets:

    def __init__(self):
        self.validated_responses = {
            'orderbook': {
                'coinbase_pro': False,
                'binance': False
            },
            'prices': {
                'coinbase_pro': False,
                'binance': False,
                'ftx':  False,
            }
        }

        kwargs = {'keys_path': './tests/config/keys.json',
                  'settings_path': "./tests/config/settings.json"}
        # Create an exchange to reference the settings.json
        blankly.CoinbasePro(**kwargs)

        self.price_manager = blankly.TickerManager('coinbase_pro', 'BTC-USD')
        self.orderbook_manager = blankly.OrderbookManager('coinbase_pro', 'BTC-USD')

        self.passed = False

        self.test_websockets()

    def test_websockets(self):

        self.price_manager.create_ticker(self.coinbase_price, override_symbol='BTC-USD',
                                         override_exchange='coinbase_pro')

        # Subscribe to a bunch with binance
        self.price_manager.create_ticker(self.binance_price, override_symbol='BTC-USDT', override_exchange='binance')
        self.price_manager.create_ticker(self.binance_price, override_symbol='ETH-USDT', override_exchange='binance')

        self.price_manager.create_ticker(self.ftx_price, override_symbol='BTC-USD', override_exchange='ftx')

        self.orderbook_manager.create_orderbook(self.coinbase_orderbook, override_symbol='BTC-USD',
                                                override_exchange='coinbase_pro')

        self.orderbook_manager.create_orderbook(self.binance_orderbook, override_symbol='BTC-USDT',
                                                override_exchange='binance')

        if self.timeout(180):
            self.passed = True

    def timeout(self, start_at: int):
        def check_responses() -> bool:
            for i in self.validated_responses['orderbook']:
                if not self.validated_responses['orderbook'][i]:
                    return False

            for i in self.validated_responses['prices']:
                if not self.validated_responses['prices'][i]:
                    return False

            return True

        while start_at > 0:
            # It can pass if it's valid within here
            if check_responses():
                assert True
                return True
            time.sleep(1)
            start_at -= 1

        # Better exit before here
        # If this is thrown then the websockets aren't returning values
        print(blankly.utils.pretty_print_json(self.validated_responses))
        assert False

    @staticmethod
    def validate_price_event(event: dict):
        try:
            assert 'symbol' in event
            assert 'price' in event
            assert 'time' in event
            assert 'trade_id' in event
            assert 'size' in event
            return True
        except AssertionError:
            return False

    @staticmethod
    def validate_orderbook_event(event: dict):
        try:
            # Validate the keys
            assert (set(event.keys()) == {'bids', 'asks'})

            # Validate the internal types
            assert (isinstance(event['bids'][0], tuple))
            assert (isinstance(event['asks'][0], tuple))

            return True
        except AssertionError:
            return False

    """
    These can async validate each response
    """
    def coinbase_price(self, message):
        if self.validate_price_event(message):
            self.validated_responses['prices']['coinbase_pro'] = True
            self.price_manager.close_websocket('BTC-USD', 'coinbase_pro')

    def binance_price(self, message):
        if self.validate_price_event(message):
            self.validated_responses['prices']['binance'] = True
            self.price_manager.close_websocket('BTC-USDT', 'binance')
            self.price_manager.close_websocket('ETH-USDT', 'binance')

    def ftx_price(self, message):
        if self.validate_price_event(message):
            self.validated_responses['prices']['ftx'] = True
            self.price_manager.close_websocket('BTC-USD', 'ftx')

    """
    These can async validate each response
    """
    def coinbase_orderbook(self, message):
        if self.validate_orderbook_event(message):
            self.validated_responses['orderbook']['coinbase_pro'] = True
            self.orderbook_manager.close_websocket('BTC-USD', 'coinbase_pro')

    def binance_orderbook(self, message):
        if self.validate_orderbook_event(message):
            self.validated_responses['orderbook']['binance'] = True
            self.orderbook_manager.close_websocket('BTC-USDT', 'binance')



"""
    Logic to provide consistency across exchanges
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import abc
from datetime import datetime as dt
from typing import Union

import pandas

from blankly.utils.utils import AttributeDict
from blankly.exchanges.orders.limit_order import LimitOrder
from blankly.exchanges.orders.market_order import MarketOrder


class ABCExchangeInterface(abc.ABC):
    @abc.abstractmethod
    def __init__(self, exchange_name, authenticated_API):
        """
        Create an abstract exchange interface
        Args:
            exchange_name (str): Define exchange name ex: 'binance' or 'coinbase_pro'
            authenticated_API (obj): Authenticated direct calls object
        """
        pass

    @abc.abstractmethod
    def get_calls(self):
        """
        Get the direct & authenticated exchange object

        Returns:
             The exchange's direct calls object. A blankly Bot class should have immediate access to this by
             default
        """
        pass

    @abc.abstractmethod
    def get_exchange_type(self):
        """
        Get the type of exchange ex: "coinbase_pro" or "binance"

        Returns:
             A string that corresponds to the type of exchange

        TODO add return example
        """
        pass

    @abc.abstractmethod
    def get_products(self) -> list:
        """
        Get all trading pairs on the exchange & some information about the exchange limits.

        TODO add return example
        """
        pass

    @abc.abstractmethod
    def get_account(self,
                    symbol: str = None) -> AttributeDict:
        """
        Get all assets in an account, or sort by symbol/account_id
        Args:
            symbol (Optional): Filter by particular symbol

            These arguments are mutually exclusive

        TODO add return example
        """
        pass

    @abc.abstractmethod
    def market_order(self,
                     symbol: str,
                     side: str,
                     size: float) -> MarketOrder:
        """
        Used for buying or selling market orders
        Args:
            symbol: asset to buy
            side: buy/sell
            size: desired amount of base asset to use
        """
        pass

    @abc.abstractmethod
    def limit_order(self,
                    symbol: str,
                    side: str,
                    price: float,
                    size: float) -> LimitOrder:
        """
        Used for buying or selling limit orders
        Args:
            symbol: asset to buy
            side: buy/sell
            price: price to set limit order
            size: amount of asset (like BTC) for the limit to be valued
        """
        pass

    @abc.abstractmethod
    def cancel_order(self,
                     symbol: str,
                     order_id: str) -> dict:
        """
        Cancel an order on a particular symbol & order id

        Args:
            symbol: This is the asset id that the order is under
            order_id: The unique ID of the order.

        TODO add return example
        """

    @abc.abstractmethod
    def get_open_orders(self,
                        symbol: str = None) -> list:
        """
        List open orders.
        Args:
            symbol (optional) (str): Asset such as BTC-USD
        TODO add return example
        """
        pass

    @abc.abstractmethod
    def get_order(self,
                  symbol: str,
                  order_id: str) -> dict:
        """
        Get a certain order
        Args:
            symbol: Asset that the order is under
            order_id: The unique ID of the order.
        TODO add return example
        """
        pass

    @abc.abstractmethod
    def get_fees(self) -> dict:
        """
        Get market fees
        TODO add return example
        """
        pass

    """
    Coinbase Pro: get_account_history
    binance: 
        get_deposit_history
        get_withdraw_history

    """

    @abc.abstractmethod
    def history(self,
                symbol: str,
                to: Union[str, int] = 200,
                resolution: Union[str, int] = '1d',
                start_date: Union[str, dt, float] = None,
                end_date: Union[str, dt, float] = None,
                return_as: str = 'df') -> pandas.DataFrame:
        """
        Wrapper for .get_product_history() which allows users to more easily get product history from right now.
        Args:
            symbol: Blankly product ID format (BTC-USD)
            to (str or int): The number of data points back in time either expressed as a string
                ("1y" meaning 1 year back") or int of points (300 points at specified resolution)
            resolution: Resolution as a string (i.e. "1d", "4h", "1y")
            start_date (str or datetime or float): Start Date for data gathering (in either string, datetime or epoch
                timestamp)
            end_date (str or datetime or float): End Date for data gathering (in either string, datetime or epoch
                timestamp)
            return_as (str): Return Type (Either list or df (dataframe))
        Returns:
            Dataframe with *at least* 'time (epoch)', 'low', 'high', 'open', 'close', 'volume' as columns.
            TODO add return example
        """
        pass

    @abc.abstractmethod
    def get_product_history(self,
                            symbol: str,
                            epoch_start: float,
                            epoch_stop: float,
                            resolution: Union[str, int]) -> pandas.DataFrame:
        """
        Returns the product history from an exchange
        Args:
            symbol: The asset such as (BTC-USD, or MSFT)
            epoch_start: Time to begin download
            epoch_stop: Time to stop download
            resolution: Resolution in seconds between tick (ex: 60 = 1 per minute)
        Returns:
            Dataframe with *at least* 'time (epoch)', 'low', 'high', 'open', 'close', 'volume' as columns.
            TODO add return example
        """

    @abc.abstractmethod
    def get_order_filter(self,
                         symbol: str) -> dict:
        """
        Find order limits for the exchange
        Args:
            symbol: The asset such as (BTC-USD, or MSFT)
            TODO add return example
        """
        pass

    @abc.abstractmethod
    def get_price(self,
                  symbol: str) -> float:
        """
        Returns just the price of a symbol.
        Args:
            symbol: The asset such as (BTC-USD, or MSFT)
            TODO add return example
        """
        pass

    @property
    @abc.abstractmethod
    def account(self) -> AttributeDict:
        """
        Get all assets in an account, or sort by assets/account_id

        TODO add return example
        """
        pass

    @property
    @abc.abstractmethod
    def orders(self) -> list:
        """
        List open orders.
        TODO add return example
        """
        pass

    @property
    @abc.abstractmethod
    def cash(self) -> float:
        """
        Get the amount of cash in a portfolio. The cash default is set in the settings .json file
        """
        pass



"""
    Backtesting controller for paper trading interface
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import traceback
import typing
from datetime import datetime as dt
import copy

import numpy as np
import pandas as pd
from bokeh.layouts import column as bokeh_columns
from bokeh.models import HoverTool
from bokeh.palettes import Category10_10
from bokeh.plotting import ColumnDataSource, figure, show

import blankly.exchanges.interfaces.paper_trade.metrics as metrics
from blankly.exchanges.interfaces.paper_trade.backtest_result import BacktestResult
from blankly.exchanges.interfaces.paper_trade.paper_trade import PaperTrade
from blankly.exchanges.interfaces.paper_trade.paper_trade_interface import PaperTradeInterface
from blankly.utils.time_builder import time_interval_to_seconds
from blankly.utils.utils import load_backtest_preferences, update_progress, write_backtest_preferences, \
    get_base_asset, get_quote_asset, info_print


def to_string_key(separated_list):
    output = ""
    for i in range(len(separated_list) - 1):
        output += str(separated_list[i])
        output += ","
    output += str(separated_list[-1:][0])
    return output


def split(base_range, local_segments) -> typing.Tuple[list, list]:
    """
    Find the negative given from a range and a set of other ranges

    Args:
        base_range: Backing array containing a range such as [1, 10]
        local_segments: An array of segments that we have such as [-5, 5], [6,7]
    Returns:
        The output of the example inputs above would be [[8, 10]]
    """

    # If we don't have any local segments there is no need for any of this, just download the whole set
    used_ranges = []
    positive_ranges = []  # These are the ranges that we have downloaded
    negative_ranges = []  # These are the ranges that we need

    def intersection(as_, ae, bs, be):
        """
        Find the intersection of two ranges
        Implementation from https://scicomp.stackexchange.com/a/26260

        Args:
            as_: Low of first range
            ae: High of first range
            bs: Low of second range
            be: High of second range
        """
        if bs > ae or as_ > be:
            return None
        else:
            os_ = max(as_, bs)
            oe = min(ae, be)
            return [os_, oe]
    for i in local_segments:
        intersection_range = intersection(base_range[0], base_range[1], i[0], i[1])
        if intersection_range is None:
            continue
        # We now know every single downloaded range that we need to keep
        used_ranges.append(i)

        # Now these are all intersections that we will filter to find the negative ranges to download
        positive_ranges.append(intersection_range)

    # Create a list of indexes that need to be pushed into a single list, arrays like  [1, 5] and [3, 7] would need to
    #  become [1, 7]

    # First we have to sort the lists to have the same starting value
    positive_ranges = sorted(positive_ranges, key=lambda x: x[0])

    aggregate_indexes = []
    for i in range(len(positive_ranges) - 1):
        intersection_range = intersection(positive_ranges[i][0], positive_ranges[i][1],
                                          positive_ranges[i + 1][0],     positive_ranges[i + 1][1])
        if intersection_range is not None:
            aggregate_indexes.append([i, i + 1])

    # Now aggregate from the last result
    for i in aggregate_indexes:
        # Pull the lower bound right out of the lower index from the aggregate_indexes approach
        lower_bound = positive_ranges[i[0]][0]

        positive_ranges[i[0]] = None

        # Push that same lower bound into the array that was above it
        positive_ranges[i[1]][0] = lower_bound

    # Filter out the None's that were added
    positive_ranges = [x for x in positive_ranges if not (x is None)]

    # Now just try to find the gaps in the positive ranges and those are our negative ranges
    for i in range(len(positive_ranges) - 1):
        negative_ranges.append([positive_ranges[i][1], positive_ranges[i+1][0]])

    try:
        # Now we just have to make sure to check the bounds
        # If the very first value we have is greater than what we requested
        if positive_ranges[0][0] > base_range[0]:
            # We then need the base range up to our first positive value
            negative_ranges.append([base_range[0], positive_ranges[0][0]])

        # If the last positive range is less than our requested final
        if positive_ranges[-1][1] < base_range[1]:
            # The actual value is from the last positive range value to our requested final value
            negative_ranges.append([positive_ranges[-1][1], base_range[1]])
    except IndexError:
        return [], [base_range]

    return used_ranges, negative_ranges


class BackTestController:
    def __init__(self, paper_trade_exchange: PaperTrade, backtest_settings_path: str = None, callbacks: list = None):
        self.preferences = load_backtest_preferences(backtest_settings_path)
        self.backtest_settings_path = backtest_settings_path
        if not paper_trade_exchange.get_type() == "paper_trade":
            raise ValueError("Backtest controller was not constructed with a paper trade exchange object.")
        self.interface: PaperTradeInterface = paper_trade_exchange.get_interface()

        self.price_events = []

        if callbacks is None:
            callbacks = []
        self.callbacks = callbacks

        self.current_time = None
        self.initial_time = None

        self.prices = []  # [epoch, "BTC-USD", price]

        self.price_dictionary = {}

        self.pd_prices = None

        # User added times
        self.__user_added_times = []

        self.use_price = None

        self.queue_backtest_write = False

        self.quote_currency = None

        # Create a global generator because a second yield function gets really nasty
        self.__color_generator = Category10_10.__iter__()

        self.initial_account = None

        self.__exchange_type = self.interface.get_exchange_type()

        # Create our own traded assets' dictionary because we customize it a bit
        self.__traded_assets = []

        # Because the times are run in order we can use this variable to optimize account value searching
        self.__current_search_index = 0

        # Export a time for use in other classes
        self.time = None

        # Set these when prices are added to find the first price and the last price
        self.user_start = None
        self.user_stop = None

        self.min_resolution = None

    def sync_prices(self) -> dict:
        """
        Parse the local file cache for the requested data, if it doesn't exist, request it from the exchange

        args:
            items: list of lists organized as ['symbol', 'start_time', 'end_time', 'resolution']

        returns:
            dictionary with keys for each 'symbol'
        """
        
        cache_folder = self.preferences['settings']["cache_location"]
        # Make sure the cache folder exists and read files
        try:
            files = os.listdir(cache_folder)
        except FileNotFoundError:
            files = []
            os.mkdir(cache_folder)

        available_files = []
        for i in range(len(files)):
            # example file name: 'BTC-USD.1622400000.1622510793.60.csv'
            # Remove the .csv from each of the files: BTC-USD.1622400000.1622510793.60
            identifier = files[i][:-4].split(",")
            # Cast to float first before
            try:
                identifier[1] = int(float(identifier[1]))
                identifier[2] = int(float(identifier[2]))
                identifier[3] = int(float(identifier[3]))
                available_files.append(identifier)
            except IndexError:
                raise IndexError(f"Please remove file {files[i]} in your price caches or "
                                 f"match the blankly cache format.")

        # This is only the downloaded data
        local_history_blocks = {}
        for i in available_files:
            # Sort these into resolutions
            asset = i[0]

            # Make sure to add the asset as a key
            if asset not in list(local_history_blocks.keys()):
                local_history_blocks[asset] = {}

            # Add each resolution to its own internal list
            if i[3] not in list(local_history_blocks[asset].keys()):
                local_history_blocks[asset][i[3]] = [i]
            else:
                local_history_blocks[asset][i[3]].append(i)

        # def segment(conjoined_ranges: list) -> list:
        #     """
        #     Split a list with sets of even intervals into multiple lists. The list must be single increments in all
        #     places except for the splits
        #     """
        #     ranges = []
        #     segments = 0
        #     for i in range(len(conjoined_ranges)-1):
        #         i += 1
        #         if conjoined_ranges[i] - conjoined_ranges[i-1] != 1:
        #             segmented_list = conjoined_ranges[segments:i]
        #
        #             ranges.append((segmented_list[0], segmented_list[-1]))
        #             segments = i + 1
        #
        #     return ranges

        # This is the data the user has requested: [asset_id, start_time, end_time, resolution]
        items = self.__user_added_times

        final_prices = {}
        for i in range(len(items)):
            if items[i] is None:
                continue
            asset = items[i][0]
            resolution = items[i][3]
            start_time = items[i][1]
            end_time = items[i][2] - resolution

            if end_time < start_time:
                raise RuntimeError("Must specify a longer timeframe to run the backtest.")

            download_ranges = []

            # Attempt to find the same symbol/asset possibilities in the backtest blocks
            try:
                # Make sure to copy it because if you don't you delete this for any similar resolution
                # If you don't copy it fails if you use two price events at the same resolution
                download_ranges = copy.deepcopy(local_history_blocks[asset][resolution])
            except KeyError:
                pass

            for j in range(len(download_ranges)):
                download_ranges[j] = download_ranges[j][1:3]

            used_ranges, negative_ranges = split([start_time, end_time], download_ranges)

            # for j in download_ranges:
            #     download_start_time = j[1]
            #     download_end_time = j[2] - resolution
            #
            #     contains_start = False
            #     contains_end = False
            #     if download_start_time <= start_time <= download_end_time:
            #         contains_start = True
            #     elif download_start_time <= end_time <= download_end_time:
            #         contains_end = True
            #
            #     if contains_start or contains_end:
            #         # Read in the whole thing if it has any sort of relevance on the data
            #         relevant_df = pd.read_csv(os.path.join(cache_folder,
            #                                                to_string_key(j) + ".csv"))
            #
            #         # If it only contains the start then it must be offset too high or is too long
            #         if contains_start and not contains_end:
            #             relevant_df = relevant_df[relevant_df['time'] >= start_time]
            #
            #             # Set the end time to match the first time in the relevant dataframe
            #             self.__user_added_times[i][1] = relevant_df['time'].iloc[-1]
            #
            #         # If it only contains the end then it must be offset too low or is too short
            #         if contains_end and not contains_start:
            #             relevant_df = relevant_df[relevant_df['time'] <= end_time]
            #
            #             # Set the start time to match the very last time of the relevant dataframe
            #             self.__user_added_times[i][2] = relevant_df['time'].iloc[0]
            #
            #         # If it contains start and end then it's just right
            #         if contains_end and contains_start:
            #             relevant_df = relevant_df[relevant_df['time'] >= start_time]
            #             relevant_df = relevant_df[relevant_df['time'] <= end_time]
            #
            #             # This dataset has been completed so there is no need to continue
            #             self.__user_added_times[i] = None
            #
            #         # Write these into the data array
            #         if asset not in list(final_prices.keys()):
            #             final_prices[asset] = relevant_df
            #         else:
            #             final_prices[asset] = pd.concat([final_prices[asset], relevant_df])

            relevant_data = []
            for j in used_ranges:
                relevant_data.append(pd.read_csv(os.path.join(cache_folder, to_string_key([asset, j[0], j[1],
                                                                                           resolution]) + ".csv")))

            if len(relevant_data) > 0:
                final_prices[asset] = pd.concat(relevant_data)

            # If there is any data left to download do it here
            for j in negative_ranges:
                print("No cached data found for " + asset + " from: " + str(j[0]) + " to " +
                      str(j[1]) + " at a resolution of " + str(resolution) + " seconds.")
                download = self.interface.get_product_history(asset,
                                                              j[0],
                                                              j[1],
                                                              resolution)

                # Write the file but this time include very accurately the start and end times
                if self.preferences['settings']['continuous_caching']:
                    if not download.empty:
                        download.to_csv(os.path.join(cache_folder, f'{asset},'
                                                                   f'{j[0]},'
                                                                   f'{j[1]+resolution},'  # This adds resolution 
                                                                                          # back to the exported
                                                                                          # time series
                                                                   f'{resolution}.csv'),
                                        index=False)

                # Write these into the data array
                if asset not in list(final_prices.keys()):
                    final_prices[asset] = download
                else:
                    final_prices[asset] = pd.concat([final_prices[asset], download])

            # After all the negative ranges are appended, we need to sort & trim
            final_prices[asset] = final_prices[asset].sort_values(by=['time'], ignore_index=True)

            # Now make sure to just trim our times to hit the start and end times
            final_prices[asset] = final_prices[asset][final_prices[asset]['time'] >= start_time]
            final_prices[asset] = final_prices[asset][final_prices[asset]['time'] <= end_time + resolution]  # Add back

        return final_prices

        # cache_folder = self.preferences['settings']["cache_location"]
        # # Make sure the cache folder exists and read files
        # try:
        #     files = os.listdir(cache_folder)
        # except FileNotFoundError:
        #     files = []
        #     os.mkdir(cache_folder)
        #
        # available_files = []
        # for i in range(len(files)):
        #     # example file name: 'BTC-USD.1622400000.1622510793.60.csv'
        #     # Remove the .csv from each of the files: BTC-USD.1622400000.1622510793.60
        #     identifier = files[i][:-4].split(",")
        #     identifier[1] = float(identifier[1])
        #     identifier[2] = float(identifier[2])
        #     identifier[3] = float(identifier[3])
        #     available_files.append(identifier)
        #
        # assets = self.preferences['price_data']['assets']  # type: dict
        # for i in assets:
        #     identifier = [i[0], i[1], i[2], i[3]]
        #     string_identifier = to_string_key(identifier)
        #     if identifier in available_files:
        #         # Read the csv here
        #         if tuple(identifier) not in self.price_dictionary.keys():
        #             print("Including: " + string_identifier + ".csv in backtest.")
        #             self.price_dictionary[tuple(identifier)] = pd.read_csv(os.path.join(cache_folder,
        #                                                                                 string_identifier + ".csv")
        #                                                                    )
        #     else:
        #         if tuple(identifier) not in self.price_dictionary.keys():
        #             print("No exact cache exists for " + str(identifier[0]) + " from " + str(identifier[1]) + " to " +
        #                   str(identifier[2]) + " at " + str(identifier[3]) + "s resolution. Downloading...")
        #             download = self.interface.get_product_history(identifier[0], identifier[1], identifier[2],
        #                                                           identifier[3])
        #             self.price_dictionary[tuple(identifier)] = download
        #             if save:
        #                 download.to_csv(os.path.join(cache_folder, string_identifier + ".csv"), index=False)
        #
        # # Merge all the same asset ids into the same dictionary spots
        # unique_assets = {}
        # for k, v in self.price_dictionary.items():
        #     if k[0] in unique_assets:
        #         unique_assets[k[0]] = pd.concat([unique_assets[k[0]], v], ignore_index=True)
        #     else:
        #         unique_assets[k[0]] = v
        #
        # return unique_assets

    def add_prices(self, asset_id, start_time, end_time, resolution, save=False):
        # Create its unique identifier
        identifier = [asset_id, int(start_time), int(end_time), int(resolution)]

        # If it's not loaded then write it to the file
        if tuple(identifier) not in self.price_dictionary.keys():
            # self.preferences['price_data']['assets'].append(identifier)
            self.__user_added_times.append(identifier)
            if save:
                self.queue_backtest_write = True

            # This makes sure that we keep track of our bounds which is just generally kind of useful
            # Any time a new price event is added we check this
            if self.user_start is None:
                self.user_start = start_time
            else:
                if start_time < self.user_start:
                    self.user_start = start_time

            if self.user_stop is None:
                self.user_stop = end_time
            else:
                if end_time > self.user_stop:
                    self.user_stop = end_time

            # Now keep track of the smallest price event added
            if self.min_resolution is None:
                self.min_resolution = resolution
            else:
                if resolution < self.min_resolution:
                    self.min_resolution = resolution
        else:
            print("already identified")

    def append_backtest_price_event(self, callback: typing.Callable, asset_id, time_interval, state_object, ohlc, init,
                                    teardown):
        if isinstance(time_interval, str):
            time_interval = time_interval_to_seconds(time_interval)
        self.price_events.append([callback, asset_id, time_interval, state_object, ohlc, init, teardown])

    def write_setting(self, key, value, save=False):
        """
        Write a setting to the .json preferences

        Args:
            key: Key under settings
            value: Value to set that settings key to
            save: Write this new setting to the file
        """
        self.preferences['settings'][key] = value
        if save:
            self.queue_backtest_write = True

    def write_initial_price_values(self, account_dictionary):
        """
        Write in a new price dictionary for the paper trade exchange.
        """
        self.interface.override_local_account(account_dictionary)

    def format_account_data(self, local_time) -> typing.Tuple[typing.Dict[typing.Union[str, typing.Any],
                                                                          typing.Union[int, typing.Any]],
                                                              typing.Dict[typing.Union[str, typing.Any],
                                                                          typing.Union[int, typing.Any]]]:

        # This is done so that only traded assets are evaluated.
        true_available = {}
        assets = self.__traded_assets
        true_account = {}
        for i in assets:
            # Grab the account status
            true_account[i] = self.interface.get_account(i)

        # Create an account total value
        value_total = 0

        no_trade_available = {}
        # No trade account total
        no_trade_value = 0

        # Save this up front so that it can be removed from the price calculation (it's always a value of 1 anyway)
        quote_value = true_account[self.quote_currency]['available'] + true_account[self.quote_currency]['hold']
        try:
            del true_account[self.quote_currency]
        except KeyError:
            pass

        for i in list(true_account.keys()):
            # Funds on hold are still added
            true_available[i] = true_account[i]['available'] + true_account[i]['hold']
            no_trade_available[i] = self.initial_account[i]['available'] + self.initial_account[i]['hold']
            currency_pair = i

            # Convert to quote (this could be optimized a bit)
            if self.__exchange_type != 'alpaca':
                currency_pair += '-'
                currency_pair += self.quote_currency

            # Get price at time
            try:
                price = self.interface.get_price(currency_pair)
            except KeyError:
                # Must be a currency we have no data for
                price = 0
            value_total += price * true_available[i]
            no_trade_value += price * no_trade_available[i]

        # Make sure to add the time key in
        true_available['time'] = local_time
        no_trade_available['time'] = local_time

        value_total += quote_value
        true_available[self.quote_currency] = quote_value

        no_trade_value += self.initial_account[self.quote_currency]['available'] + \
                          self.initial_account[self.quote_currency]['hold']

        true_available['Account Value (' + self.quote_currency + ')'] = value_total

        no_trade_available['Account Value (No Trades)'] = no_trade_value

        return true_available, no_trade_available

    def __account_was_used(self, column) -> bool:
        show_zero_delta = self.preferences['settings']['show_tickers_with_zero_delta']

        # Just check if it's in the traded assets or if the zero delta is enabled
        used = self.__traded_assets
        is_used = column in used or 'Account Value (' + self.quote_currency + ')' == column

        # Return true if they are not the same or the setting is set to true
        output = is_used or show_zero_delta
        return output

    def __next_color(self):
        # This should be a generator, but it doesn't work without doing a foreach loop
        try:
            return next(self.__color_generator)
        except StopIteration:
            self.__color_generator = Category10_10.__iter__()
            return next(self.__color_generator)

    def run(self) -> BacktestResult:
        """
        Setup
        """

        # Create this initial so that we can compare how our strategy performs
        self.initial_account = self.interface.get_account()

        # Write our queued edits to the file
        if self.queue_backtest_write:
            write_backtest_preferences(self.preferences, self.backtest_settings_path)

        # TODO the preferences should really be reloaded here so that micro changes such as the quote currency reset
        #  don't need to happen for every single key type
        self.quote_currency = self.preferences['settings']['quote_account_value_in']

        # Get the symbol used for the benchmark
        benchmark_symbol = self.preferences["settings"]["benchmark_symbol"]

        if benchmark_symbol is not None:
            # Check locally for the data and add to price_cache if we do not have it
            self.add_prices(benchmark_symbol, self.user_start, self.user_stop, self.min_resolution)

        prices = self.sync_prices()

        # Organize each price into this structure: [epoch, "BTC-USD", price, open, high, low, close, volume]
        use_price = self.preferences['settings']['use_price']
        self.use_price = use_price

        self.pd_prices = {**prices}

        for k, v in prices.items():
            frame = v  # type: pd.DataFrame

            # Be sure to push these initial prices to the strategy
            try:
                self.interface.receive_price(k, v[use_price].iloc[0])
            except IndexError:
                def check_if_any_column_has_prices(price_dict: dict) -> bool:
                    """
                    In dictionary of symbols, check if at least one key has data
                    """
                    for j in price_dict:
                        if not price_dict[j].empty:
                            return True
                    return False

                if not check_if_any_column_has_prices(prices):
                    raise IndexError('No cached or downloaded data available. Try adding arguments such as to="1y" '
                                     'in the backtest command. If there should be data downloaded, try deleting your'
                                     ' ./price_caches folder.')
                else:
                    raise IndexError(f"Data for symbol {k} is empty. Are you using a symbol that is incompatible "
                                     f"with this exchange?")

            # Be sure to send in the initial time
            self.interface.receive_time(v['time'].iloc[0])

            for index, row in frame.iterrows():
                # TODO iterrows() is allegedly pretty slow
                self.prices.append([row.time, k, row[use_price],
                                    row['open'],  # (index) 3
                                    row['high'],  # 4
                                    row['low'],  # 5
                                    row['close'],  # 6
                                    row['volume']])  # 7

        # pushing these prices together makes the time go weird
        self.prices = sorted(self.prices)

        if prices == {} or self.price_events == []:
            raise ValueError("Either no price data or backtest events given. "
                             "Try setting an argument such as to='1y' in the .backtest() command.\n"
                             "Example: strategy.backtest(to='1y')")

        self.current_time = self.prices[0][0]

        self.initial_time = self.current_time

        # Add a section to the price events which controls the next time they run & change to array of dicts
        for i in range(len(self.price_events)):
            self.price_events[i] = {
                'function': self.price_events[i][0],
                'asset_id': self.price_events[i][1],
                'interval': self.price_events[i][2],
                'state_object': self.price_events[i][3],
                'next_run': self.initial_time,
                'ohlc': self.price_events[i][4],
                'init': self.price_events[i][5],
                'teardown': self.price_events[i][6]
            }

        # Initialize this before the callbacks so it works in the initialization functions
        self.time = self.initial_time

        # Turn on backtesting immediately after setting the time
        self.interface.set_backtesting(True)

        # Run the initialization functions for the price events
        print("\nInitializing...")
        for i in self.price_events:
            if i['init'] is not None:
                i['init'](i['asset_id'], i['state_object'])

        """
        Begin backtesting
        """

        # Re-evaluate the traded assets account
        # This is mainly used if the user has an account with some value that gets added in at the backtest point
        # This occurs after initialization so there has to be a function to test & re-evaluate that
        self.interface.evaluate_traded_account_assets()
        column_keys = list.copy(self.interface.traded_assets)

        # Comically if you don't include the quote at any point there will be an error
        if self.quote_currency not in column_keys:
            column_keys.append(self.quote_currency)
        # If they start a price event on something they don't own, this should also be included
        for i in self.price_events:
            base_asset = get_base_asset(i['asset_id'])
            quote_asset = get_quote_asset(i['asset_id'])
            if base_asset not in column_keys:
                column_keys.append(base_asset)
            if quote_asset not in column_keys:
                column_keys.append(quote_asset)
        self.__traded_assets = list.copy(column_keys)
        column_keys.append('time')

        # column_keys = ['time']
        # for i in account.keys():
        #     column_keys.append(i['currency'])

        cycle_status = pd.DataFrame(columns=column_keys)

        no_trade_cycle_status = pd.DataFrame(columns=column_keys)

        # Append dictionaries to this to make the pandas dataframe
        price_data = []

        # Append dictionaries to this to make the no trade dataframe
        no_trade = []

        # Add an initial account row here
        if self.preferences['settings']['save_initial_account_value']:
            available_dict, no_trade_dict = self.format_account_data(self.initial_time)
            price_data.append(available_dict)
            no_trade.append(no_trade_dict)

        show_progress = self.preferences['settings']['show_progress_during_backtest']

        ignore_exceptions = self.preferences['settings']['ignore_user_exceptions']

        print("\nBacktesting...")
        price_number = len(self.prices)

        # This dictionary will hold price arrays below sorted by symbol which will allow us to grab most
        #  recent ohlcv data when necessary
        ohlcv_tracker = {}

        try:
            for i in range(price_number):
                #                 row.time,      k,  use_price, 'open', 'high', 'low','close','volume'
                # Formatted like [1609146000.0, 'AAPL', 132.99, 132.72, 133.0, 132.6, 132.99, 32603.0]
                price_array = self.prices[i]
                if show_progress:
                    if i % 100 == 0:
                        update_progress(i / price_number)
                self.interface.receive_price(asset_id=price_array[1], new_price=price_array[2])
                self.current_time = price_array[0]
                self.interface.evaluate_limits()

                # This will keep the most recent price event data organized by symbol
                ohlcv_tracker[price_array[1]] = price_array

                while True:
                    # Need to go through and establish an order for each of the price events
                    self.price_events = sorted(self.price_events, key=lambda sort_key: sort_key['next_run'])

                    # Now the lowest one has to go past the current time to be invalid
                    if self.price_events[0]['next_run'] > self.current_time:
                        break

                    local_time = self.price_events[0]['next_run']

                    # Export the time for strategy
                    self.time = local_time

                    self.interface.receive_time(local_time)

                    # This is the actual callback to the user space
                    try:
                        if self.price_events[0]['ohlc']:
                            # This pulls all the price data out of the price array defined on line 260
                            ohlcv_array = ohlcv_tracker[self.price_events[0]['asset_id']]
                            self.price_events[0]['function']({'open': ohlcv_array[3],
                                                              'high': ohlcv_array[4],
                                                              'low': ohlcv_array[5],
                                                              'close': ohlcv_array[6],
                                                              'volume': ohlcv_array[7]},

                                                             self.price_events[0]['asset_id'],
                                                             self.price_events[0]['state_object'])
                        else:
                            self.price_events[0]['function'](self.interface.get_price(self.price_events[0]['asset_id']),
                                                             self.price_events[0]['asset_id'], self.price_events[0][
                                                                 'state_object'])
                    except Exception as e:
                        if ignore_exceptions:
                            traceback.print_exc()
                        else:
                            raise e

                    # Delay the next run until after the interval
                    self.price_events[0]['next_run'] += self.price_events[0]['interval']

                    available_dict, no_trade_dict = self.format_account_data(local_time)

                    price_data.append(available_dict)

                    no_trade.append(no_trade_dict)

            # Finally, run the teardown functions
            for i in self.price_events:
                # Pull the teardown and pass the state object
                if callable(i['teardown']):
                    i['teardown'](i['state_object'])
        except Exception:
            traceback.print_exc()

        # Reset time to be None to indicate we're no longer in a backtest
        self.time = None

        # Push the accounts to the dataframe
        cycle_status = cycle_status.append(price_data, ignore_index=True).sort_values(by=['time'])

        if len(cycle_status) == 0:
            raise RuntimeError("Empty result - no valid backtesting events occurred. Was there an error?.")

        no_trade_cycle_status = no_trade_cycle_status.append(no_trade, ignore_index=True).sort_values(by=['time'])

        figures = []
        # for i in self.prices:
        #     result_index = cycle_status['time'].sub(i[0]).abs().idxmin()
        #     for i in cycle_status.iloc[result_index]:

        hover = HoverTool(
            tooltips=[
                ('value', '@value')
            ],

            # formatters={
            #     'time': 'datetime',  # use 'datetime' formatter for 'date' field
            #     '@{value}': 'printf',   # use 'printf' formatter for '@{adj close}' field
            # },

            # display a tooltip whenever the cursor is vertically in line with a glyph
            mode='vline'
        )
        
        # Define a helper function to avoid repeating code
        def add_trace(self_, figure_, time_, data_, label):
            source = ColumnDataSource(data=dict(
                        time=time_,
                        value=data_.values.tolist()
                    ))
            figure_.step('time', 'value',
                         source=source,
                         line_width=2,
                         color=self_.__next_color(),
                         legend_label=label,
                         mode="after")

        if self.preferences['settings']['GUI_output']:
            global_x_range = None

            time = [dt.fromtimestamp(ts) for ts in cycle_status['time']]

            for column in cycle_status:
                if column != 'time' and self.__account_was_used(column):
                    p = figure(plot_width=900, plot_height=200, x_axis_type='datetime')
                    add_trace(self, p, time, cycle_status[column], column)

                    # Add the no-trade line to the backtest
                    if column == 'Account Value (' + self.quote_currency + ')':
                        add_trace(self, p, time, no_trade_cycle_status['Account Value (No Trades)'],
                                  'Account Value (No Trades)')

                        # Add the benchmark, if requested
                        if benchmark_symbol is not None:
                            # This normalizes the benchmark value
                            initial_account_value = cycle_status['Account Value (' + self.quote_currency + ')'].iloc[0]
                            initial_benchmark_value = prices[benchmark_symbol][use_price].iloc[0]

                            # This multiplier brings the initial asset price to the initial account value
                            # initial_account_value = initial_benchmark_value * x
                            multiplier = initial_account_value / initial_benchmark_value

                            normalized_compare_series = prices[benchmark_symbol][use_price].multiply(multiplier)
                            normalized_compare_time_series = prices[benchmark_symbol]['time']

                            # We need to also cast the time series that is needed to compare
                            # because it's only been done for the cycle status time
                            normalized_compare_time_series = [dt.fromtimestamp(ts) for ts in
                                                              normalized_compare_time_series]
                            add_trace(self, p, normalized_compare_time_series,
                                      normalized_compare_series,
                                      f'Normalized Benchmark ({benchmark_symbol})')
                            
                    p.add_tools(hover)

                    # Format graph
                    p.legend.location = "top_left"
                    p.legend.title = column
                    p.legend.title_text_font_style = "bold"
                    p.legend.title_text_font_size = "20px"
                    if global_x_range is None:
                        global_x_range = p.x_range
                    else:
                        p.x_range = global_x_range

                    figures.append(p)

            show(bokeh_columns(figures))

        def is_number(s):
            try:
                float(s)
                # Love how bools cast to a number
                return not isinstance(s, bool)
            except ValueError:
                return False

        history_and_returns = {
            'history': cycle_status
        }
        metrics_indicators = {}
        user_callbacks = {}

        result_object = BacktestResult(history_and_returns, {
            'created': self.interface.paper_trade_orders,
            'limits_executed': self.interface.executed_orders,
            'limits_canceled': self.interface.canceled_orders,
            'executed_market_orders': self.interface.market_order_execution_details
        }, self.pd_prices, self.initial_time, self.interface.time(), self.quote_currency, self.price_events, figures)

        # If they set resampling we use resampling for everything
        resample_setting = self.preferences['settings']['resample_account_value_for_metrics']
        if isinstance(resample_setting, str) or is_number(resample_setting):
            resample_to = resample_setting
        else:
            info_print('Resampling value not set, defaulting to 1 day.')
            resample_to = '1d'

        interval_value = time_interval_to_seconds(resample_to)

        # This is where we run the actual resample
        resampled_account_data_frame = result_object.resample_account('Account Value (' + self.quote_currency + ')',
                                                                      interval_value)

        history_and_returns['resampled_account_value'] = resampled_account_data_frame

        returns = resampled_account_data_frame.copy(deep=True)

        # Default diff parameters should do it
        returns['value'] = returns['value'].pct_change()

        # Now write it to our dictionary
        history_and_returns['returns'] = returns

        # -----=====*****=====-----
        metrics_indicators['Compound Annual Growth Rate (%)'] = metrics.cagr(history_and_returns)
        try:
            metrics_indicators['Cumulative Returns (%)'] = metrics.cum_returns(history_and_returns)
        except ZeroDivisionError:
            raise ZeroDivisionError("Division by zero when calculating cumulative returns. "
                                    "Are there valid account datapoints?")

        def attempt(math_callable: typing.Callable, dict_of_dataframes: dict, kwargs: dict = None):
            try:
                if kwargs is None:
                    kwargs = {}
                result = math_callable(dict_of_dataframes, **kwargs)
                if result == np.NAN:
                    result = None
                return result
            except (ZeroDivisionError, Exception) as e_:
                return f'failed: {e_}'

        risk_free_return_rate = self.preferences['settings']["risk_free_return_rate"]
        metrics_indicators['Max Drawdown (%)'] = attempt(metrics.max_drawdown, history_and_returns)
        metrics_indicators['Variance (%)'] = attempt(metrics.variance, history_and_returns,
                                                     {'trading_period': interval_value})
        metrics_indicators['Sortino Ratio'] = attempt(metrics.sortino, history_and_returns,
                                                      {'risk_free_rate': risk_free_return_rate,
                                                       'trading_period': interval_value})
        metrics_indicators['Sharpe Ratio'] = attempt(metrics.sharpe, history_and_returns,
                                                     {'risk_free_rate': risk_free_return_rate,
                                                      'trading_period': interval_value})
        metrics_indicators['Calmar Ratio'] = attempt(metrics.calmar, history_and_returns, 
                                                     {'trading_period': interval_value})
        metrics_indicators['Volatility'] = attempt(metrics.volatility, history_and_returns,
                                                   {'trading_period': interval_value})
        metrics_indicators['Value-at-Risk'] = attempt(metrics.var, history_and_returns)
        metrics_indicators['Conditional Value-at-Risk'] = attempt(metrics.cvar, history_and_returns)
        
        # Add risk-free-return rate to dictionary
        metrics_indicators['Risk Free Return Rate'] = risk_free_return_rate
        # metrics_indicators['beta'] = attempt(metrics.beta, dataframes)
        # Add the interval value to dictionary
        metrics_indicators['Resampled Time'] = interval_value
        # -----=====*****=====-----

        # If a benchmark was requested, add it to the pd_prices frame
        if benchmark_symbol is not None:
            # Resample the benchmark results
            resampled_benchmark_value = result_object.resample_account(benchmark_symbol,
                                                                       interval_value,
                                                                       use_asset_history=True,
                                                                       use_price=use_price)
            
            # Push data into the dictionary for use by the metrics
            history_and_returns['benchmark_value'] = resampled_benchmark_value
            history_and_returns['benchmark_returns'] = resampled_benchmark_value.copy(deep=True)
            history_and_returns['benchmark_returns']['value'] = \
                history_and_returns['benchmark_returns']['value'].pct_change()
            
            # Calculate beta
            metrics_indicators['Beta'] = attempt(metrics.beta, history_and_returns, 
                                                 {"trading_period": interval_value})

        # This trys to reference vars created in the resample portion, so it has to also be in the specified
        #  resample if
        # --
        # Run this last so that the user can override what they want
        for callback in self.callbacks:
            user_callbacks[callback.__name__] = callback(history_and_returns, metrics_indicators)

        # Remove NaN values here
        history_and_returns['resampled_account_value'] = history_and_returns['resampled_account_value'].\
            where(history_and_returns['resampled_account_value'].notnull(), None)

        # Remove NaN values on this one too
        history_and_returns['returns'] = history_and_returns['returns'].where(history_and_returns['returns'].notnull(),
                                                                              None)
        # Lastly remove Nan values in the metrics
        for i in metrics_indicators:
            if not isinstance(metrics_indicators[i], str) and np.isnan(metrics_indicators[i]):
                metrics_indicators[i] = None

        # Assign all these new values back to the result object
        result_object.history_and_returns = history_and_returns
        result_object.metrics = metrics_indicators
        result_object.user_callbacks = user_callbacks

        self.interface.set_backtesting(False)
        return result_object



"""
    This shows a simple golden cross example on MSFT, checking at a 1d resolution over two years
"""

from blankly import Strategy, StrategyState, Interface
from blankly import Alpaca
from blankly.indicators import sma


def init(symbol, state: StrategyState):
    interface: Interface = state.interface
    resolution = state.resolution
    variables = state.variables
    # initialize the historical data
    variables['history'] = interface.history(symbol, 1200, resolution, end_date=state.time, return_as='deque')['close']
    variables['owns_position'] = False


def price_event(price, symbol, state: StrategyState):
    interface: Interface = state.interface
    variables = state.variables

    variables['history'].append(price)

    sma200 = sma(variables['history'], period=200)
    # match up dimensions
    sma50 = sma(variables['history'], period=50)[-len(sma200):]
    diff = sma200 - sma50
    slope_sma50 = (sma50[-1] - sma50[-5]) / 5  # get the slope of the last 5 SMA50 Data Points
    prev_diff = diff[-2]
    curr_diff = diff[-1]
    is_cross_up = slope_sma50 > 0 and curr_diff >= 0 and prev_diff < 0
    is_cross_down = slope_sma50 < 0 and curr_diff <= 0 and prev_diff > 0
    # comparing prev diff with current diff will show a cross
    if is_cross_up and not variables['owns_position']:
        interface.market_order(symbol, 'buy', int(interface.cash/price))
        variables['owns_position'] = True
    elif is_cross_down and variables['owns_position']:
        # use strategy.base_asset if on CoinbasePro or Binance
        interface.market_order(symbol, 'sell', int(interface.account[symbol].available))
        variables['owns_position'] = False


if __name__ == "__main__":
    alpaca = Alpaca()
    s = Strategy(alpaca)
    s.add_price_event(price_event, 'SNAP', resolution='1h', init=init)
    s.add_price_event(price_event, 'GME', resolution='1h', init=init)
    s.backtest(initial_values={'USD': 10000}, to='2y')
    # Or just run it directly on the exchange
    # s.start()



"""
    Statistics tests
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pickle
import unittest
from pathlib import Path

import numpy as np

from blankly.indicators import max_period, min_period, stddev_period, stderr_period, var_period


def compare_equal(a, b):
    # compares two numpy arrays
    return np.array_equal(a, b)


class Statistics(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        data_path = Path("tests/config/test_data.p").resolve()
        with open(data_path, 'rb') as f:
            cls.data = pickle.load(f)

    def test_one_param_moving_averages(self):
        period = self.data['periods'][0]

        stddev_res = stddev_period(self.data['close'], period)
        self.assertTrue(compare_equal(stddev_res, self.data[period]['stddev_period']))

        var_res = var_period(self.data['close'], period)
        self.assertTrue(compare_equal(var_res, self.data[period]['var_period']))

        stderr_res = stderr_period(self.data['close'], period)
        self.assertTrue(compare_equal(stderr_res, self.data[period]['stderr_period']))

        min_res = min_period(self.data['close'], period)
        self.assertTrue(compare_equal(min_res, self.data[period]['min_period']))

        max_res = max_period(self.data['close'], period)
        self.assertTrue(compare_equal(max_res, self.data[period]['max_period']))

    def test_one_param_moving_averages_diff_period(self):
        period = self.data['periods'][1]

        stddev_res = stddev_period(self.data['close'], period)
        self.assertTrue(compare_equal(stddev_res, self.data[period]['stddev_period']))

        var_res = var_period(self.data['close'], period)
        self.assertTrue(compare_equal(var_res, self.data[period]['var_period']))

        stderr_res = stderr_period(self.data['close'], period)
        self.assertTrue(compare_equal(stderr_res, self.data[period]['stderr_period']))

        min_res = min_period(self.data['close'], period)
        self.assertTrue(compare_equal(min_res, self.data[period]['min_period']))

        max_res = max_period(self.data['close'], period)
        self.assertTrue(compare_equal(max_res, self.data[period]['max_period']))


"""
    Oscillator tests
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import pickle
import unittest
from pathlib import Path

import numpy as np

from blankly.indicators import absolute_price_oscillator, aroon_oscillator, chande_momentum_oscillator, \
    percentage_price_oscillator, rsi, stochastic_oscillator


def compare_equal(a, b):
    # compares two numpy arrays
    return np.array_equal(a, b)


class Oscillators(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        data_path = Path("tests/config/test_data.p").resolve()
        with open(data_path, 'rb') as f:
            cls.data = pickle.load(f)

    def test_rsi(self):
        for period in self.data['periods']:
            rsi_res = rsi(self.data['close'], period)
            self.assertTrue(compare_equal(rsi_res, self.data[period]['rsi']))

    def test_aroon_oscillator(self):
        for period in self.data['periods']:
            aroon_res = aroon_oscillator(self.data['high'], self.data['low'], period)
            self.assertTrue(compare_equal(aroon_res, self.data[period]['aroon_oscillator']))

    def test_chande_momentum_oscillator(self):
        for period in self.data['periods']:
            chande_res = chande_momentum_oscillator(self.data['close'], period)
            self.assertTrue(compare_equal(chande_res, self.data[period]['chande_momentum_oscillator']))


    def test_absolute_price_oscillator(self):
        short_period = self.data['short_period']
        long_period = self.data['long_period']
        res = absolute_price_oscillator(self.data['close'], short_period, long_period)
        # use the first one since general period doesn't matter
        self.assertTrue(compare_equal(res, self.data[self.data['periods'][0]]['absolute_price_oscillator']))

    def test_percentage_price_oscillator(self):
        short_period = self.data['short_period']
        long_period = self.data['long_period']
        res = percentage_price_oscillator(self.data['close'], short_period, long_period)
        # use the first one since general period doesn't matter
        self.assertTrue(compare_equal(res, self.data[self.data['periods'][0]]['percentage_price_oscillator']))

    def test_stochastic_oscillator(self):
        pct_k_period = self.data['pct_k_period']
        pct_k_slowing_period = self.data['pct_k_slowing_period']
        pct_d_period = self.data['pct_d_period']
        res = stochastic_oscillator(self.data['high'], self.data['low'], self.data['close'], pct_k_period, pct_k_slowing_period, pct_d_period)
        # use the first one since general period doesn't matter
        self.assertTrue(compare_equal(res, self.data[self.data['periods'][0]]['stochastic_oscillator']))


"""
    Custom alpaca unit tests
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import datetime
import time
from datetime import datetime as dt
from pathlib import Path

import alpaca_trade_api
import dateparser
import dateparser as dp
import pytest
import pytz

import blankly
from blankly.exchanges.interfaces.alpaca.alpaca_interface import AlpacaInterface

timeZ_Ny = pytz.timezone('America/New_York')
MARKET_OPEN = datetime.time(hour=9, minute=0, second=0, tzinfo=timeZ_Ny)
MARKET_CLOSE = datetime.time(hour=17, minute=0, second=0, tzinfo=timeZ_Ny)


@pytest.fixture
def alpaca_interface():
    keys_file_path = Path("tests/config/keys.json").resolve()
    settings_file_path = Path("tests/config/settings.json").resolve()

    alpaca = blankly.Alpaca(keys_path=keys_file_path,
                            settings_path=settings_file_path,
                            portfolio_name="alpaca test portfolio")

    # auth_obj = AlpacaAuth(str(keys_file_path), "alpaca test portfolio")
    # _, alpaca_interface = DirectCallsFactory.create("alpaca", auth_obj, str(settings_file_path))
    return alpaca.interface


def test_get_exchange(alpaca_interface: AlpacaInterface) -> None:
    assert alpaca_interface.get_exchange_type() == 'alpaca'


# TODO: uncomment and fix this when we get the function to work
# def test_get_products(alpaca_interface: AlpacaInterface) -> None:
#     return_val = alpaca_interface.get_products()
#
#     expected_answer = {
#         "symbol": "AAPL-USD",
#         "base_asset": "AAPL",
#         "quote_asset": "USD",
#         "base_min_size": -1,
#         "base_max_size": -1,
#         "base_increment": -1,
#     }
#     for k, v in expected_answer.items():
#         assert return_val[0][k] == v
#
#     assert "exchange_specific" in return_val[0]

def test_get_buy_sell(alpaca_interface: AlpacaInterface) -> None:
    if not alpaca_interface.get_calls().get_clock()['is_open']:
        return

    known_apple_info = {
        "id": "904837e3-3b76-47ec-b432-046db621571b",
        "class": "us_equity",
        "exchange": "NASDAQ",
        "symbol": "AAPL",
        "status": "active",
        "tradable": True,
        "marginable": True,
        "shortable": True,
        "easy_to_borrow": True,
        "fractionable": True
    }

    # query for the unique ID of AAPL
    products = alpaca_interface.get_products()
    apple_id = None

    for product in products:
        if product['base_asset'] == 'AAPL':
            apple_id = product['exchange_specific']['id']

    assert apple_id

    market_order = alpaca_interface.market_order(apple_id, 'buy', 1)

    num_retries = 0
    # wait until order is filled
    while alpaca_interface.get_order('AAPL', market_order.get_id())['size'] is None:
        time.sleep(1)
        num_retries += 1
        if num_retries > 10:
            raise TimeoutError("Too many retries, cannot get order status")

    resp = alpaca_interface.get_order('AAPL', market_order.get_id())

    # get_position(symbol)

    # get_asset(symbol)

    # place sell order


def test_get_product_history(alpaca_interface: AlpacaInterface) -> None:
    start = dt.strptime("2021-02-04", "%Y-%m-%d").timestamp()
    end = dt.strptime("2021-02-05", "%Y-%m-%d").timestamp()

    return_df = alpaca_interface.get_product_history("AAPL", start, end, 60)
    return_df_2 = alpaca_interface.get_product_history("AAPL", start, end, 120)
    assert len(return_df) > 0
    assert len(return_df) >= len(return_df_2)


def test_get_product_history_est_timezone(alpaca_interface: AlpacaInterface) -> None:
    start = dateparser.parse("2021-02-04 9:30AM EST").timestamp()
    end = dateparser.parse("2021-02-04 9:35AM EST").timestamp()

    return_df = alpaca_interface.get_product_history("AAPL", start, end, 60)
    desired_timestamp = dp.parse("2021-02-04 14:30:00+00:00").timestamp()
    assert return_df.iloc[0]['time'] == desired_timestamp
    assert (len(return_df) == 6)


def test_get_product_history_custom(alpaca_interface: AlpacaInterface) -> None:
    assert isinstance(alpaca_interface.calls, alpaca_trade_api.REST)
    end = dateparser.parse("2021-02-04 9:35AM EST")

    return_df = alpaca_interface.history("AAPL", to=10, resolution='1h', end_date=end)
    assert (len(return_df) == 10)


def test_get_account(alpaca_interface: AlpacaInterface) -> None:
    products = alpaca_interface.get_account()
    for _, val in products.items():
        assert 'available' in val
        assert 'hold' in val

    assert "USD" in products


def test_get_price(alpaca_interface: AlpacaInterface) -> None:
    price = alpaca_interface.get_price("AAPL")
    assert isinstance(price, float)

def test_get_open_orders(alpaca_interface: AlpacaInterface) -> None:
    orders = alpaca_interface.get_open_orders()
    print(orders)



"""
    Inherited authentication object
    Copyright (C) 2021  Arun Annamalai, Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import abc

from blankly.exchanges.auth.utils import load_auth
from blankly.utils.exceptions import AuthError
from blankly.utils.utils import info_print


class AuthConstructor(abc.ABC):
    def __init__(self, keys_file: str, portfolio_name: str, exchange: str, needed_keys: list):
        """
        Create an auth interface
        Args:
            keys_file (str): filepath to keys.json
            portfolio_name (str): name of portfolio
            exchange (str): name of exchange
        """
        # self.portfolio_name = portfolio_name
        self.exchange = exchange

        # Load from file
        self.portfolio_name, self.keys = load_auth(exchange, keys_file, portfolio_name)

        self.validate_credentials(needed_keys)

    def validate_credentials(self, needed_keys: list):
        """
        Args:
            needed_keys (list): List of keys that the exchange needs in the format ['API_KEY', 'API_SECRET]

        Returns:
            A boolean that describes if the auth is good or bad (True if good).
        """

        # Create an error message template to throw if needed
        error_message = ""

        # Make a copy of the keys dict & list to avoid modifying passed variables
        keys_dict = {**self.keys}
        needed_keys = list.copy(needed_keys)
        for i in needed_keys:
            if i in keys_dict.keys():
                keys_dict.pop(i)
            else:
                # Append a description header if not present
                if error_message == "":
                    error_message += "Error while loading authentication. Required keys for this are missing: \n"
                error_message += str(str(i) + " is needed, but not defined.\n")

        if len(keys_dict.keys()) > 0:
            info_print(f"Additional keys for Exchange: {self.exchange} Portfolio: {self.portfolio_name} will be"
                       f" ignored.")

        if error_message != "":
            raise AuthError(error_message)



"""

ATTENTION:
FTX has NO sandbox mode, these are methods specifically designed
to test interface methods which interact with LIVE orders, and 
move REAL funds. It is not recommended that you run these tests.
They must be executed individually and with great caution. Blankly
is not responsible for losses incurred by improper use of this test. 

How do we test methods which affect live funds?

There are 5 main methods that need to be tested which involve
using live funds

1. get_open_orders()
2. market_order()
3. limit_order()
4. cancel_order

We test these by creating a series of assertations that can only
pass if everything works properly. 

"""

import time
import sys

import blankly
from pathlib import Path
from typing import List

from blankly.exchanges.interfaces.ftx.ftx_interface import FTXInterface
import test_ftx_interface_utils as test_utils


def ftx_interface() -> FTXInterface:
    keys_file_path = Path("../../../config/keys.json").resolve()
    settings_file_path = Path("../../../config/settings.json").resolve()

    ftx = blankly.FTX(keys_path=keys_file_path,
                               settings_path=settings_file_path,
                               portfolio_name='Main Account')

    # auth_obj = FTXAuth(str(keys_file_path), "Main Account")
    # _, ftx_interface = DirectCallsFactory.create("ftx", auth_obj, str(settings_file_path))
    return ftx.interface


"""
cancels ALL open orders and distributes funds for testing purposes

pass command line argument reset to activate

use with extreme caution

"""


def reset(ftx_interface: FTXInterface):
    validation = input(
        "Continuing with this test will cancel ALL existing open orders and redistribute funds for testing purposes. "
        "Continue (Y/n)? ")
    if validation.lower() != "y":
        print("aborting test")
        quit()

    for order in ftx_interface.get_open_orders():
        ftx_interface.cancel_order("BTC/USD", order["id"])


def main():
    interface: FTXInterface = ftx_interface()

    validation = input(
        "To complete this test properly, you should have around $10 in USD in your FTX account.\nThis test will "
        "involve the conversion of no more than $10 to BTC.\nDo not have any trades going on while completing this "
        "test.\nContinue (Y/n)? ")
    if validation.lower() != "y":
        quit()

    if len(sys.argv) == 2:
        if sys.argv[1] == "reset":
            reset(interface)
        else:
            print("Invalid command line argument. Aborting")

    # start testing ###

    initial_num_open_orders = len(interface.get_open_orders())

    # creates a buy limit order unlikely to be fulfilled (buying a bitcoin for $5)
    limit_order = interface.limit_order("BTC/USD", "buy", 1, 5)
    assert len(
        interface.get_open_orders()) == 1 + initial_num_open_orders, f"expected {1 + initial_num_open_orders} " \
                                                                     f"but got {len(interface.get_open_orders())}"
    assert limit_order.get_status()['status'] == "new" or limit_order.get_status()[
        'status'] == "open", f"expected open but got {limit_order.get_status()['status']} of " \
                             f"type {type(limit_order.get_status()['status'])}"

    open_order_ids: List[int] = test_utils.get_open_order_ids(interface.get_open_orders())
    limit_order_id = limit_order.get_id()
    assert limit_order_id in open_order_ids, f"id {limit_order_id} (type: {type(limit_order_id)}) not in " \
                                             f"list {open_order_ids} (type: {type(open_order_ids)})"

    # cancels the open limit order
    order_id = int(interface.cancel_order("BTC/USD", limit_order.get_id()))

    # wait for order cancellation to go through
    print("waiting for order cancellation to be accepted")
    while interface.get_order("BTC/USD", order_id)["status"] != "closed":
        print(
            f"waiting for order to be cancelled. Current status: " + interface.get_order("BTC/USD", order_id)["status"],
            end='\r')
        time.sleep(1)
    time.sleep(5)
    assert (interface.get_order("BTC/USD", order_id)["status"] == "closed")
    print("order successfully cancelled")

    curr_num_open_orders = len(interface.get_open_orders())
    assert curr_num_open_orders == initial_num_open_orders, f"expected {initial_num_open_orders} " \
                                                            f"open orders, but got {curr_num_open_orders}\nopen " \
                                                            f"order list:\n{interface.get_open_orders()}"
    assert order_id not in test_utils.get_open_order_ids(interface.get_open_orders())

    # creates a buy market order for $5 worth of bitcoin
    market_order = interface.market_order("BTC/USD", "buy", 5 / interface.get_price("BTC/USD"))

    curr_order_status = ""
    # waits for status of market order to update
    while curr_order_status != "closed":
        curr_order_status = interface.get_order("BTC/USD", market_order.get_id())["status"]
        print("waiting for test market order to get accepted. Current status: " + curr_order_status, end='\r')
        time.sleep(1)

    time.sleep(5)

    print("Test market order accepted... waiting for status to update")

    market_order_info = interface.get_order("BTC/USD", market_order.get_id())

    assert market_order.get_id() not in test_utils.get_open_order_ids(interface.get_open_orders())
    assert market_order_info["id"] == market_order.get_id()
    market_order_status = market_order_info["status"]
    assert market_order_status == "closed", f"expected closed, but got {market_order_status} (type: " \
                                            f"{type(market_order_status)}) \nopen orders: {interface.get_open_orders()}"

    # creates a sell limit order unlikely to get fulfilled (selling ~$5 worth of bitcoin for $50,000)
    price = 50000
    size = 5 / interface.get_price("BTC/USD")
    print(f"Attempting to create a limit order of {size} BTC for {price} USD")
    limit_order = interface.limit_order("BTC/USD", "sell", price, size)

    curr_open_orders = test_utils.get_open_order_ids(interface.get_open_orders())
    assert limit_order.get_id() in curr_open_orders, \
        f"expected order id {limit_order.get_id()} to be in list of open orders: \n{curr_open_orders}"
    limit_order_info = interface.get_order("BTC/USD", limit_order.get_id())
    assert limit_order_info["id"] == limit_order.get_id()

    print("Waiting for limit order status to update. Please sit tight...")

    curr_order_status = limit_order_info["status"]

    while curr_order_status == "new":
        curr_order_status = interface.get_order("BTC/USD", limit_order.get_id())["status"]
        print(f"Waiting on order status to get updated. Current order status: {curr_order_status}", end="\r")
        time.sleep(1)
    time.sleep(5)
    print("Limit order successfully closed")

    assert curr_order_status == "open", f"expected open but got {curr_order_status} (type: {type(curr_order_status)})"
    assert len(interface.get_open_orders()) == initial_num_open_orders + 1

    # cancels sell limit order
    order_id = interface.cancel_order("BTC/USD", limit_order.get_id())
    assert order_id not in test_utils.get_open_order_ids(interface.get_open_orders())
    assert len(interface.get_open_orders()) == initial_num_open_orders

    # creates sell market order
    market_order = interface.market_order("BTC/USD", "sell", 5 / interface.get_price("BTC/USD"))

    while market_order.get_status() == "open":
        print("Waiting for test market order to get accepted...", end='\r')
        time.sleep(1)

    time.sleep(5)

    print("Test market order accepted... continuing test")

    market_order_info = interface.get_order("BTC/USD", market_order.get_id())

    assert market_order.get_id() not in test_utils.get_open_order_ids(interface.get_open_orders())
    assert market_order_info["id"] == market_order.get_id()
    assert market_order_info["status"] == "closed", f"expected closed but got" + market_order_info["status"]

    # end testing ###

    print("\n --- All tests passed --- \n")


if __name__ == "__main__":

    """    
    Feel free to run the test as many times in a row as you would like for redundancy purposes
    
    net change in distribution of funds as a result of testing, between before and after the test,
    is appoximately zero
    """

    for i in range(1):
        main()



"""
    Unit tests for testing the strategy behaviors
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import unittest

import blankly
from blankly import StrategyState
from blankly.utils.utils import get_quote_asset

from tests.testing_utils import get_valid_symbol


class StrategyTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.ran_price_event = False
        cls.ran_bar_event = False

    def price_event(self, price, symbol, state: StrategyState):
        self.assertTrue(isinstance(price, float))

        filter_ = state.interface.get_order_filter(symbol)

        min_size = filter_['market_order']['base_min_size']

        # Ensure you check for ints
        if min_size == 1.0:
            min_size = 1

        state.interface.market_order(symbol, 'buy', min_size)

        state.interface.market_order(symbol, 'buy', min_size)

        self.ran_price_event = True

    def bar_event(self, price, symbol, state: StrategyState):
        self.assertTrue(isinstance(price, dict))
        # Check ohlcv on bar
        self.assertTrue(set(price.keys()) == {'open', 'high', 'low', 'close', 'volume'})
        self.ran_bar_event = True

    def test_backtest(self):
        kwargs = {'keys_path': './tests/config/keys.json',
                  'settings_path': "./tests/config/settings.json"}

        self.exchanges = [
            blankly.Kucoin(**kwargs),
            blankly.FTX(**kwargs),
            blankly.CoinbasePro(**kwargs),
            blankly.Binance(**kwargs),
            blankly.Oanda(**kwargs),
            blankly.Alpaca(**kwargs)
        ]

        for i in range(len(self.exchanges)):
            strategy = blankly.Strategy(self.exchanges[i])
            type_ = self.exchanges[i].interface.get_exchange_type()
            strategy.add_price_event(self.price_event, get_valid_symbol(type_), '1d')
            strategy.add_bar_event(self.bar_event, get_valid_symbol(type_), '1d')
            print(f'Associated {get_valid_symbol(type_)} with {type_}.')

            print(f"Testing: {self.exchanges[i].interface.get_exchange_type()} without benchmark")
            quote_asset = get_quote_asset(get_valid_symbol(self.exchanges[i].interface.get_exchange_type()))
            strategy.backtest(to='1y', settings_path='./tests/config/backtest.json',
                              quote_account_value_in=quote_asset,
                              initial_values={quote_asset: 1000000000},
                              benchmark_symbol=None,
                              continuous_caching=False)
            # Of course, it'll pass if the function never runs, make sure it does
            assert self.ran_price_event
            self.ran_price_event = False
            assert self.ran_bar_event
            self.ran_bar_event = False
            print(f"Testing: {self.exchanges[i].interface.get_exchange_type()} with benchmark")
            strategy.backtest(to='1y', settings_path='./tests/config/backtest.json',
                              quote_account_value_in=get_quote_asset(get_valid_symbol(
                                        self.exchanges[i].interface.get_exchange_type())),
                              benchmark_symbol=get_valid_symbol(self.exchanges[i].interface.get_exchange_type()),
                              initial_values={quote_asset: 1000000000},
                              continuous_caching=False)
            assert self.ran_price_event
            self.ran_price_event = False
            assert self.ran_bar_event
            self.ran_bar_event = False



"""
    General FTX Api definition
    Copyright (C) 2021 Blankly Finance

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import requests
from typing import Optional, Dict, Any, List
import urllib.parse
from blankly.utils.utils import epoch_from_iso8601
import time
import hmac


class FTXAPI:
    _API_URL = "https://ftx.us/api/"

    # no option to instantiate with sandbox mode, unlike every other exchange
    def __init__(self, api_key, api_secret, _subaccount_name=None):

        self._ftx_session = requests.Session()
        self._api_key = api_key
        self._api_secret = api_secret
        self._subaccount_name = _subaccount_name

    def _signed_request(self, method: str, path: str, **kwargs):
        request = requests.Request(method, self._API_URL + path, **kwargs)
        self._get_signature(request)
        result = self._ftx_session.send(request.prepare())
        return self._handle_response(result)

    def _get_signature(self, request: requests.Request):
        ts = int(1000 * time.time())
        prepared_request = request.prepare()
        signed_data = f'{ts}{prepared_request.method}{prepared_request.path_url}'.encode()

        if prepared_request.body:
            signed_data += prepared_request.body

        signature = hmac.new(self._api_secret.encode(), signed_data, 'sha256').hexdigest()

        request.headers['FTXUS-KEY'] = self._api_key
        request.headers['FTXUS-SIGN'] = signature
        request.headers['FTXUS-TS'] = str(ts)

        if self._subaccount_name:
            request.headers['FTXUS-SUBACCOUNT'] = urllib.parse.quote(self._subaccount_name)

    def _signed_delete(self, path: str, params: Optional[Dict[str, Any]] = None):
        return self._signed_request('DELETE', path, json=params)

    def _signed_get(self, path: str, params: Optional[Dict[str, Any]] = None):
        return self._signed_request('GET', path, params=params)

    def _signed_post(self, path: str, params: Optional[Dict[str, Any]] = None):
        return self._signed_request('POST', path, json=params)

    @staticmethod
    def _handle_response(response: requests.Response):
        try:
            result = response.json()

        except ValueError:
            response.raise_for_status()
            raise

        else:
            if not result['success']:
                raise Exception(result['error'])
            else:
                return result['result']

    def list_markets(self) -> List[dict]:

        return self._signed_get('markets')

    def get_market(self, symbol: str) -> dict:

        # symbol format must be BTC/USD instead of BTC-USD
        if '-' in symbol:
            symbol = symbol.replace("-", "/")

        return self._signed_get(f'markets/{symbol}')

    def list_futures(self) -> List[dict]:
        return self._signed_get('futures')

    def get_product_history(self, symbol: str, epoch_start: float, epoch_stop: float, resolution) -> List[dict]:
        return self._signed_get(
            f"/markets/{symbol}/candles?resolution={resolution}&start_time={epoch_start}&end_time={epoch_stop}")

    def get_orderbook(self, market: str, depth: int = None) -> dict:
        return self._signed_get(f'markets/{market}/orderbook', {'depth': depth})

    def get_trades(self, market: str) -> dict:
        return self._signed_get(f'markets/{market}/trades')

    def get_account_info(self) -> dict:
        return self._signed_get('account')

    def get_open_orders(self, market: str = None) -> List[dict]:
        return self._signed_get('orders', {'market': market})

    def get_order_by_id(self, order_id: str) -> dict:
        return self._signed_get(f'orders/{order_id}')

    def get_fills(self) -> List[dict]:
        return self._signed_get('fills')

    def get_balances(self) -> List[dict]:
        return self._signed_get('wallet/balances')

    def get_deposit_addresses(self, ticker: str) -> dict:
        return self._signed_get(f'wallet/deposit_addresses/{ticker}')

    def get_positions(self, display_price_avg: bool = False) -> List[dict]:
        return self._signed_get('positions', {'showAvgPrice': display_price_avg})

    def get_specific_position(self, pos_name: str, display_price_avg: bool = False) -> dict:
        filtered = filter(lambda pos: pos['future'] == pos_name, self.get_positions(display_price_avg))
        return next(filtered, None)

    def get_order_history(
            self,
            market: str = None,
            side: str = None,
            order_type: str = None,
            interval_start: float = None,
            interval_end: float = None
    ) -> List[dict]:
        return self._signed_get('orders/history', {
            'market': market,
            'side': side,
            'orderType': order_type,
            'start_time': interval_start,
            'end_time': interval_end})

    def get_conditional_order_history(
            self,
            market: str = None,
            side: str = None,
            type_: str = None,
            order_type: str = None,
            interval_start: float = None,
            interval_end: float = None
    ) -> List[dict]:
        return self._signed_get(f'conditional_orders/history', {
            'market': market,
            'side': side,
            'type': type_,
            'orderType': order_type,
            'start_time': interval_start,
            'end_time': interval_end
        })

    def get_conditional_orders(self, market: str = None) -> List[dict]:
        return self._signed_get(f'conditional_orders', {'market': market})

    def place_order(
            self,
            market: str,
            side: str,
            price: float,
            size: float,
            order_type: str = 'limit',
            reduce_only: bool = False,
            ioc: bool = False,
            post_only: bool = False,
            client_id: str = None
    ) -> dict:
        return self._signed_post('orders', {
            'market': market,
            'side': side,
            'price': price,
            'size': size,
            'type': order_type,
            'reduceOnly': reduce_only,
            'ioc': ioc,
            'postOnly': post_only,
            'clientId': client_id
        })

    def cancel_order(self, order_id: str) -> dict:
        return self._signed_delete(f'orders/{order_id}')

    """
        MUST SPECIFY EXACTLY 1 OF THE FOLLOWING
        order_id XOR existing_client_order_id
    """

    def modify_order(
            self,
            order_id: Optional[str] = None,
            existing_client_order_id: Optional[str] = None,
            client_order_id: Optional[float] = None,
            price: Optional[float] = None,
            size: Optional[float] = None,
    ) -> dict:

        assert (price is None) or (size is None), \
            'You must modify the price or the size'

        assert (existing_client_order_id is None) ^ (order_id is None), \
            'Specify exactly one ID to modify'

        if order_id:
            endpoint_path = f"orders/{order_id}/modify"
        elif existing_client_order_id:
            endpoint_path = f"orders/by_client_id/{existing_client_order_id}/modify"
        else:
            raise AttributeError('Either client_order_id or existing_client_order_id is required')

        if size is None:
            size_json = {}
        else:
            size_json = {'size': size}

        if price is None:
            price_json = {}
        else:
            price_json = {'price': price}

        if client_order_id is None:
            client_order_id_json = {}
        else:
            client_order_id_json = {'clientId': client_order_id}

        return self._signed_post(endpoint_path, {
            **size_json,
            **price_json,
            **client_order_id_json
        })

    """
    to send...
    stop market order: set type = "stop" and specify a trigger price
    stop limit order: also supply a limit price
    take profit market order: set type = "trailing_stop" and specify a trigger price
    """

    def place_conditional_order(
            self,
            market: str,
            side: str,
            size: float,
            order_type: str = "stop",
            limit_price: float = "None",
            reduce_only: bool = False,
            cancel_limit_on_trigger: bool = True,
            trigger_price: float = None,
            trail_value: float = None
    ) -> dict:
        assert type in ['stop', 'take_profit', 'trailing_stop']
        assert type not in ['stop',
                            'take_profit'] or trigger_price is not None, "stop losses and take profits requires a " \
                                                                         "trigger price"
        assert type not in ['trailing stop'] or (
                    trail_value is not None and trigger_price is not None), "must specify trail value for trailing " \
                                                                            "stop, cannot use trigger price"
        return self._signed_post(
            'conditional_orders',
            {
                'market': market,
                'side': side,
                'triggerPrice': trigger_price,
                'size': size,
                'reduceOnly': reduce_only,
                'type': order_type,
                'cancelLimitOnTrigger': cancel_limit_on_trigger,
                'orderPrice': limit_price
            }
        )

    def get_all_trades(self, market: str, interval_start: float = None, interval_end: float = None, lim=100) -> List:

        to_return = []
        set_of_ids = set()

        while True:
            resp = self._signed_get(f'markets/{market}/trades', {
                'end_time': interval_end,
                'start_time': interval_start
            })

            dd_trades = []

            for elem in resp:
                if elem['id'] not in set_of_ids:
                    dd_trades.append(elem)

            to_return.extend(dd_trades)

            set_of_ids |= {elem['id'] for elem in dd_trades}

            if len(resp) == 0:
                break
            interval_end = min(epoch_from_iso8601(time_['time']) for time_ in resp)
            if lim > len(resp):
                break

        return to_return



"""
    Indicator tests
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import pickle
import unittest
from pathlib import Path

import numpy as np

from blankly.indicators import ema, hma, kaufman_adaptive_ma, macd, sma, trima, vwma, wma, zlema


def compare_equal(a, b):
    # compares two numpy arrays
    return np.array_equal(a, b)


class MovingAverages(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        data_path = Path("tests/config/test_data.p").resolve()
        with open(data_path, 'rb') as f:
            cls.data = pickle.load(f)
        
    def test_one_param_moving_averages(self):
        period = self.data['periods'][0]

        ema_res = ema(self.data['close'], period)
        self.assertTrue(compare_equal(ema_res, self.data[period]['ema']))

        wma_res = wma(self.data['close'], period)
        self.assertTrue(compare_equal(wma_res, self.data[period]['wma']))

        zlema_res = zlema(self.data['close'], period)
        self.assertTrue(compare_equal(zlema_res, self.data[period]['zlema']))

        sma_res = sma(self.data['close'], period)
        self.assertTrue(compare_equal(sma_res, self.data[period]['sma']))

        hma_res = hma(self.data['close'], period)
        self.assertTrue(compare_equal(hma_res, self.data[period]['hma']))

        kaufman_res = kaufman_adaptive_ma(self.data['close'], period)
        self.assertTrue(compare_equal(kaufman_res, self.data[period]['kaufman_adaptive_ma']))

        trima_res = trima(self.data['close'], period)
        self.assertTrue(compare_equal(trima_res, self.data[period]['trima']))

    def test_one_param_moving_averages_diff_period(self):
        period = self.data['periods'][1]

        ema_res = ema(self.data['close'], period)
        self.assertTrue(compare_equal(ema_res, self.data[period]['ema']))

        wma_res = wma(self.data['close'], period)
        self.assertTrue(compare_equal(wma_res, self.data[period]['wma']))

        zlema_res = zlema(self.data['close'], period)
        self.assertTrue(compare_equal(zlema_res, self.data[period]['zlema']))

        sma_res = sma(self.data['close'], period)
        self.assertTrue(compare_equal(sma_res, self.data[period]['sma']))

        hma_res = hma(self.data['close'], period)
        self.assertTrue(compare_equal(hma_res, self.data[period]['hma']))

        kaufman_res = kaufman_adaptive_ma(self.data['close'], period)
        self.assertTrue(compare_equal(kaufman_res, self.data[period]['kaufman_adaptive_ma']))

        trima_res = trima(self.data['close'], period)
        self.assertTrue(compare_equal(trima_res, self.data[period]['trima']))

    def test_macd(self):
        short_period = self.data['short_period']
        long_period = self.data['long_period']
        for period in self.data['periods']:
            macd_res = macd(self.data['close'], short_period, long_period, period)
            self.assertTrue(compare_equal(macd_res, self.data[period]['macd']))

    def tests_vwma(self):
        volume = self.data['volume']
        for period in self.data['periods']:
            vwma_res = vwma(self.data['close'], volume, period)
            self.assertTrue(compare_equal(vwma_res, self.data[period]['vwma']))


"""
    Logic to provide consistency across exchanges
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import abc
import time
import warnings
from datetime import datetime as dt
from typing import Union
from collections import deque

import numpy
import pandas as pd
from dateutil import parser

import blankly.utils.utils as utils
from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface
from blankly.utils.time_builder import time_interval_to_seconds


# TODO: need to add a cancel all orders function
class ExchangeInterface(ABCExchangeInterface, abc.ABC):
    def __init__(self, exchange_name, authenticated_api, preferences_path=None, valid_resolutions=None):
        self.exchange_name = exchange_name
        self.calls = authenticated_api
        self.valid_resolutions = valid_resolutions
        # Reload user preferences here
        self.user_preferences = utils.load_user_preferences(preferences_path)

        self.exchange_properties = None
        # Some exchanges like binance will not return a value of 0.00 if there is no balance
        self.available_currencies = {}

        self.needed = {
            '__init_exchange__': [
                ['maker_fee_rate', float],
                ['taker_fee_rate', float]
            ],
            'get_products': [
                ["symbol", str],
                ["base_asset", str],
                ["quote_asset", str],
                ["base_min_size", float],
                ["base_max_size", float],
                ["base_increment", float]
            ],
            'get_account': [
                ["available", float],
                ["hold", float]
            ],
            'market_order': [
                ["symbol", str],
                ["id", str],
                ["created_at", float],
                ["size", float],
                ["status", str],
                ["type", str],
                ["side", str]
            ],
            'limit_order': [
                ["symbol", str],
                ["id", str],
                ["created_at", float],
                ["price", float],
                ["size", float],
                ["status", str],
                ["time_in_force", str],
                ["type", str],
                ["side", str]
            ],
            'stop_limit': [
                ["symbol", str],
                ["id", str],
                ["created_at", float],
                ["stop_price", float],
                ["limit_price", float],
                ["stop", str],
                ["size", float],
                ["status", str],
                ["time_in_force", str],
                ["type", str],
                ["side", str]
            ],
            'cancel_order': [
                ['order_id', str]
            ],
            # 'get_open_orders': [  # Key specificity changes based on order type
            #     ["id", str],
            #     ["price", float],
            #     ["size", float],
            #     ["type", str],
            #     ["side", str],
            #     ["status", str],
            #     ["product_id", str]
            # ],
            # 'get_order': [
            #     ["product_id", str],
            #     ["id", str],
            #     ["price", float],
            #     ["size", float],
            #     ["type", str],
            #     ["side", str],
            #     ["status", str],
            #     ["funds", float]
            # ],
            'get_fees': [
                ['maker_fee_rate', float],
                ['taker_fee_rate', float]
            ],
            'get_order_filter': [
                ["symbol", str],
                ["base_asset", str],
                ["quote_asset", str],
                ["max_orders", int],
                ["limit_order", dict],
                ["market_order", dict]
            ]
        }

        if self.user_preferences['settings']['test_connectivity_on_auth']:
            self.init_exchange()

    @abc.abstractmethod
    def init_exchange(self):
        """
        Create the properties for the exchange.

        This is never run if test_connectivity_on_auth is set to false
        """
        pass

    """ Needs to be overridden here """

    def get_calls(self):
        """
        Returns:
             The exchange's direct calls object. A blankly Bot class should have immediate access to this by
             default
        """
        return self.calls

    """ Needs to be overridden here """

    def get_exchange_type(self):
        return self.exchange_name

    @property
    def account(self):
        return utils.AttributeDict(self.get_account())

    @property
    def orders(self):
        return self.get_open_orders()

    @property
    def cash(self):
        using_setting = self.user_preferences['settings'][self.exchange_name]['cash']
        return self.get_account(using_setting)['available']

    def history(self,
                symbol: str,
                to: Union[str, int] = 200,
                resolution: Union[str, float] = '1d',
                start_date: Union[str, dt, float] = None,
                end_date: Union[str, dt, float] = None,
                return_as: str = 'df'):

        is_backtesting = self.is_backtesting()
        if is_backtesting is not None and end_date is None:
            # is_backtesting can only return a non None value if a function overrides the is_backtesting function
            #  and says its backtesting by returning a valid time
            end_date = is_backtesting

        if start_date is not None and end_date is not None:
            to = None

        to_present = False
        if end_date is None:
            to_present = True

        # convert resolution into epoch seconds
        resolution_seconds = int(time_interval_to_seconds(resolution))

        if end_date is None:
            # Figure out the next point and then subtract to the last stamp
            most_recent_valid_resolution = utils.ceil_date(dt.now(),
                                                           seconds=resolution_seconds).timestamp() - resolution_seconds
            # Binance is nice enough to update OHLCV data, so we have to exclude that by subtracting a resolution
            epoch_stop = most_recent_valid_resolution - resolution_seconds
            count_from = most_recent_valid_resolution
        else:
            if isinstance(end_date, str):
                parsed_date = parser.parse(end_date)
            elif isinstance(end_date, float) or isinstance(end_date, numpy.int64) or isinstance(end_date, int) or \
                    isinstance(end_date, numpy.int32):
                parsed_date = dt.fromtimestamp(end_date)
            else:
                parsed_date = end_date
            valid_time_in_past = utils.ceil_date(parsed_date,
                                                 seconds=resolution_seconds).timestamp() - resolution_seconds
            epoch_stop = valid_time_in_past - resolution_seconds
            count_from = valid_time_in_past

        if start_date is None and end_date is None:
            if isinstance(to, int):
                # use number of points to calculate the start epoch
                epoch_start = count_from - (to * resolution_seconds)
            else:
                epoch_start = count_from - time_interval_to_seconds(to)
        elif start_date is None and end_date is not None:
            if isinstance(to, int):
                epoch_start = count_from - (to * resolution_seconds)
            else:
                epoch_start = count_from - time_interval_to_seconds(to)
        else:
            epoch_start = utils.convert_input_to_epoch(start_date)

        response = self.overridden_history(symbol, epoch_start, epoch_stop, resolution_seconds, to=to,)

        # Add a check to make sure that coinbase pro has updated
        if to_present and self.get_exchange_type() == "coinbase_pro":
            data_append = None
            tries = 0
            while True:
                if data_append is None:
                    # We can continue if this is valid
                    if response['time'].iloc[-1] == epoch_stop:
                        break
                else:
                    if data_append[0]['time'] == epoch_stop:
                        break
                time.sleep(.5)
                tries += 1
                if tries > 5:
                    # Admit failure and return
                    warnings.warn("Exchange failed to provide updated data within the timeout.")
                    return self.cast_type(response, return_as)
                try:
                    data_append = [self.get_product_history(symbol,
                                                            epoch_stop - resolution_seconds,
                                                            epoch_stop,
                                                            resolution_seconds).iloc[-1].to_dict()]
                    data_append[0]['time'] = int(data_append[0]['time'])
                except IndexError:
                    # If there is no recent data on the exchange this will be an empty dataframe.
                    # This happens for low volume
                    utils.info_print("Most recent bar at this resolution does not yet exist - skipping.")
                    break

            response = response.append(data_append, ignore_index=True)

        # Determine the deque length - we really should use this generally
        if isinstance(to, int):
            point_count = to
        else:
            point_count = (epoch_stop-epoch_start)/resolution_seconds + 1
        return self.cast_type(response, return_as, point_count)

    def overridden_history(self, symbol, epoch_start, epoch_stop, resolution, **kwargs) -> pd.DataFrame:
        return self.get_product_history(symbol, epoch_start, epoch_stop, resolution)

    def is_backtesting(self):
        """
        Overridden by interfaces which have a valid backtesting boolean value
        """
        return None

    @staticmethod
    def evaluate_multiples(valid_resolutions: list, resolution_seconds: float):
        found_multiple = -1
        for multiple in reversed(valid_resolutions):
            if resolution_seconds % multiple == 0:
                found_multiple = multiple
                break
        if found_multiple < 0:
            raise ValueError("This exchange currently does not support this specific resolution, try making it a "
                             "multiple of 1 minute, 1 hour or 1 day")

        row_divisor = resolution_seconds / found_multiple
        row_divisor = int(row_divisor)

        if row_divisor > 100:
            raise Warning("The resolution you requested is an extremely high of the base resolutions supported and may "
                          "slow down the performance of your model: {} * {}".format(found_multiple, row_divisor))

        return found_multiple, row_divisor

    @staticmethod
    def cast_type(response: pd.DataFrame, return_as: str, point_count=None):
        if return_as != 'df' and return_as != 'deque':
            return response.to_dict(return_as)
        elif return_as == 'deque':
            # Create a deque object that has the same length
            response = response.to_dict('list')
            for i in response.keys():
                response[i] = deque(response[i], point_count)
            return response
        elif return_as == 'df':
            return response
        else:
            utils.info_print(f"Return type {return_as} is not supported.")
        return response

    def get_account(self, symbol=None):
        """
        This method is a helper that allows the get_account functions to assume the base asset is being passed in
        Args:
            symbol (Optional): Filter by particular symbol

        Coinbase Pro: get_account
        binance: get_account["balances"]
        """

        if symbol is not None:
            symbol = utils.get_base_asset(symbol)

        return symbol

    @abc.abstractmethod
    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        pass

    def choose_order_specificity(self, order_type):
        # This lower should not be necessary if everything is truly homogeneous
        order_type = order_type.lower()
        if order_type == 'market':
            return self.needed['market_order']
        elif order_type == 'limit':
            return self.needed['limit_order']
        else:
            return self.needed['market_order']

    """
    Order lifecycle should be:
    Accepted -> live -> done -> settled
    """

    @staticmethod
    def homogenize_order_status(exchange, status):
        if exchange == "binance":
            if status == "new":
                return "open"
        elif exchange == 'alpaca':
            if status == 'accepted':
                return 'accepted'
            if status == 'new':
                return 'new'

        return status
